/**
 * User layer for high-level read/write access in calimero.
 *  
 * Published under Quick'n'Dirty Licence ;)
 * So please don't use it in production code...
 * 
 * 2006 B.Malinowsky
 */

package tuwien.auto.eicl;

import java.net.InetSocketAddress;

import tuwien.auto.eibxlator.PointPDUXlator;
import tuwien.auto.eibxlator.PointPDUXlator_8BitUnsigned;
import tuwien.auto.eibxlator.PointPDUXlator_Boolean;
import tuwien.auto.eicl.event.DisconnectEvent;
import tuwien.auto.eicl.event.EICLEventListener;
import tuwien.auto.eicl.event.FrameEvent;
import tuwien.auto.eicl.struct.cemi.CEMI_L_DATA;
import tuwien.auto.eicl.struct.cemi.EIB_Address;
import tuwien.auto.eicl.util.EICLException;
import tuwien.auto.eicl.util.EICLReadTimeoutException;

/**
 * Implements high-level read/write access of data point values. It only
 * supports KNX group addresses. For now it allows reading and writing of
 * boolean values (mapped to Boolean.DPT_SWITCH), and 8BitUnsigned values
 * (mapped to 8BitUnsigned.DPT_VALUE_1_UCOUNT) and 8BitUnsigned values (mapped
 * to 8BitUnsigned.DPT_SCALING). All error conditions are handled by an
 * EICLException.
 */
public class UserLayer
{
	/**
	 * do nothing default ctor
	 */
	public UserLayer()
	{
	}

	/**
	 * set up user layer with a connection to host at port (e.g.
	 * UserLayer("fourier.auto.tuwien.ac.at", 0); )
	 * 
	 * @param host
	 *            host to connect to
	 * @param port
	 *            port at host, or 0 for default port 3671
	 * @throws EICLException
	 */
	public UserLayer(String host, int port) throws EICLException
	{
		connect(host, port);
	}

	/**
	 * set up user layer with existing open connection
	 * 
	 * @param conn
	 *            existing connection to use
	 */
	public UserLayer(CEMI_Connection conn)
	{
		if (conn == null)
			throw new NullPointerException("User layer: connection argument is null");
		this.conn = conn;
		this.conn.addFrameListener(new FrameEventListener(this));
		setConnected(true);
	}

	/**
	 * @return the read timeout used in read requests
	 */
	public int getReadTimeout()
	{
		return readTimeout;
	}

	/**
	 * set the read timeout to wait for a read call to complete
	 * 
	 * @param millis
	 *            time in milliseconds
	 */
	public void setReadTimeout(int millis)
	{
		readTimeout = millis;
	}

	public synchronized boolean isConnected()
	{
		return connected;
	}

	/**
	 * read a boolean value from group address a
	 * 
	 * @param a
	 *            group address
	 * @param result
	 *            boolean result
	 * @throws EICLException
	 */
	public Boolean readBool(EIB_Address a) throws EICLException
	{
		Boolean result;
		validState(a);

		PointPDUXlator_Boolean x = new PointPDUXlator_Boolean(PointPDUXlator_Boolean.DPT_SWITCH[0]);
		x.setServiceType(PointPDUXlator.A_GROUPVALUE_READ);
		CEMI_L_DATA data = new CEMI_L_DATA((byte) CEMI_L_DATA.MC_L_DATAREQ, new EIB_Address(), a, x.getAPDUByteArray());
		conn.sendFrame(data, CEMI_Connection.WAIT_FOR_CONFIRM);
		waitForReply(a, x);
		x.setAPDUByteArray(data.getData());
		result = x.getASDUasBoolean();

		System.out.println("Userlayer boolean read result: " + result);

		return result;
	}

	/**
	 * write a boolean to group address a
	 * 
	 * @param a
	 *            group address
	 * @param value
	 *            boolean to write
	 * @throws EICLException
	 */
	public void write(EIB_Address a, boolean value) throws EICLException
	{
		validState(a);

		PointPDUXlator_Boolean x = new PointPDUXlator_Boolean(PointPDUXlator_Boolean.DPT_SWITCH[0]);
		x.setServiceType(PointPDUXlator.A_GROUPVALUE_WRITE);
		x.setASDUfromBoolean(value);
		CEMI_L_DATA data = new CEMI_L_DATA((byte) CEMI_L_DATA.MC_L_DATAREQ, new EIB_Address(), a, x.getAPDUByteArray());
		conn.sendFrame(data, CEMI_Connection.WAIT_FOR_CONFIRM);
	}

	/**
	 * read a 8BitUnsigned (ucount or scaled) value from group address a
	 * 
	 * @param a
	 *            group address
	 * @param result
	 *            8BitUnsigned result
	 * @param scaled
	 *            true to read a scaled value, false to read ucount
	 * @throws EICLException
	 */
	public Integer readInt(EIB_Address a, boolean scaled) throws EICLException
	{
		validState(a);

		PointPDUXlator_8BitUnsigned x = new PointPDUXlator_8BitUnsigned(scaled ? PointPDUXlator_8BitUnsigned.DPT_SCALING[0] : PointPDUXlator_8BitUnsigned.DPT_VALUE_1_UCOUNT[0]);
		x.setServiceType(PointPDUXlator.A_GROUPVALUE_READ);
		CEMI_L_DATA data = new CEMI_L_DATA((byte) CEMI_L_DATA.MC_L_DATAREQ, new EIB_Address(), a, x.getAPDUByteArray());
		conn.sendFrame(data, CEMI_Connection.WAIT_FOR_CONFIRM);
		waitForReply(a, x);
		x.setAPDUByteArray(data.getData());
		return (scaled ? (int) x.getASDUasShort() : (int) x.getASDUasShortUnscaled());
	}

	/**
	 * write a 8BitUnsigned ucount or scaled value to group address a
	 * 
	 * @param a
	 *            group address
	 * @param value
	 *            8BitUnsigned to write
	 * @param scaled
	 *            true to write a scaled value, false to write ucount
	 * @throws EICLException
	 */
	public void write(EIB_Address a, int value, boolean scaled) throws EICLException
	{
		validState(a);

		PointPDUXlator_8BitUnsigned x = new PointPDUXlator_8BitUnsigned(scaled ? PointPDUXlator_8BitUnsigned.DPT_SCALING[0] : PointPDUXlator_8BitUnsigned.DPT_VALUE_1_UCOUNT[0]);
		x.setServiceType(PointPDUXlator.A_GROUPVALUE_WRITE);
		if (scaled)
			x.setASDUfromShort((short) value);
		else
			x.setASDUfromShortUnscaled((short) value);

		CEMI_L_DATA data = new CEMI_L_DATA((byte) CEMI_L_DATA.MC_L_DATAREQ, new EIB_Address(), a, x.getAPDUByteArray());
		conn.sendFrame(data, CEMI_Connection.WAIT_FOR_CONFIRM);
	}

	/**
	 * try to connect to host at port (e.g. connect("fourier.auto.tuwien.ac.at",
	 * 0); )
	 * 
	 * @param host
	 *            host address to connect (in format accepted by
	 *            InetSocketAddress)
	 * @param port
	 *            port to use or 0 for default port (3671)
	 * @throws EICLException
	 */
	public void connect(String host, int port) throws EICLException
	{
		disconnect();
		try
		{
			if (port == 0)
				port = 3671;
			InetSocketAddress a = new InetSocketAddress(host, port);
			conn = new CEMI_Connection(a, new TunnellingConnectionType());
			conn.addFrameListener(new FrameEventListener(this));
			setConnected(true);
		}
		catch (Exception e)
		{
			throw new EICLException("User layer connection error: " + e.getMessage());
		}
	}

	/**
	 * break up an existing connection
	 */
	public void disconnect()
	{
		if (conn != null)
		{
			try
			{
				setConnected(false);
				conn.disconnect("User layer disconnect");
			}
			catch (EICLException ex)
			{
				// on this level a disconnect error is not of interest
			}
		}
	}

	/**
	 * Handler for a new frame, notified through FrameEventListener
	 * 
	 * @param e
	 *            event object
	 */
	private void frameReceived(FrameEvent e)
	{
		CEMI_L_DATA data = (CEMI_L_DATA) e.getPacket();
		setCEMI(data);
		System.out.println("network: packet received from " + data.getSourceAddress().toString());
	}

	/**
	 * wait for read reply, used in read(...)
	 * 
	 * @param addr
	 *            group address to check incoming frames
	 * @param x
	 *            translator to use for parsing values
	 * @throws EICLException
	 */
	private void waitForReply(EIB_Address addr, PointPDUXlator x) throws EICLException
	{
		final int time = 5;
		setCEMI(null);
		long start = System.currentTimeMillis();
		while (true)
		{
			CEMI_L_DATA data = getCEMI();
			if (data != null && data.getMessageCode() == CEMI_L_DATA.MC_L_DATAIND)
			{
				EIB_Address a = data.getDestinationAddress();
				x.setAPDUByteArray(data.getData());
				if (a.toString().equals(addr.toString()))
				{
					// System.out.println("Service type: " +
					// x.getServiceType());
					if (x.getServiceType() == 64) // PointPDUXlator.A_GROUPVALUE_RESPONSE)
						return;
				}
				setCEMI(null);
			}
			try
			{
				Thread.sleep(time);
			}
			catch (InterruptedException e)
			{
			}
			if (System.currentTimeMillis() > start + readTimeout)
			{
				System.err.println("User layer: reply timeout after " + readTimeout + " ms");
				return;
				// throw new EICLReadTimeoutException("User layer: reply timeout
				// after " + readTimeout + " ms");
			}
		}
	}

	/**
	 * Check if object has a valid state.
	 * 
	 * @param a
	 *            EIB_Address to check against null
	 */
	private void validState(EIB_Address a)
	{
		if (conn == null)
			throw new IllegalStateException("User layer: no connection was set up");
		if (a == null)
			throw new IllegalArgumentException("User layer: address argument is null");
	}

	private synchronized void setCEMI(CEMI_L_DATA data)
	{
		cemi = data;
	}

	private synchronized CEMI_L_DATA getCEMI()
	{
		return cemi;
	}

	private synchronized void setConnected(boolean c)
	{
		connected = c;
	}

	private class FrameEventListener implements EICLEventListener
	{

		private UserLayer _callee;

		public FrameEventListener(UserLayer callee)
		{
			_callee = callee;
		}

		public void newFrameReceived(FrameEvent e)
		{
			_callee.frameReceived(e);
		}

		public void serverDisconnected(DisconnectEvent e)
		{
			_callee.connected = false;
		}
	}

	private CEMI_L_DATA cemi;

	private CEMI_Connection conn;

	private boolean connected = false;

	// maximum wait time [ms] for response in a read() call
	private int readTimeout = 10000;
}
