package tuwien.auto.obix.objectbroker;

import obix.*;
import java.io.*;

/**
 * <p>
 * The Object Broker acts as dynamic frontend to the static XML-object database.
 * If a GCE is requested or put, data from the static database is used to
 * communicate with the network and sometimes enriched with data from the
 * network and subsequently returned. Hence a 'dynamic' frontend.
 * <p>
 * 
 * @author Matthias Neugschwandtner
 */

public interface IObjectBroker
{
	/**
	 * The URI is looked up in the local database. If it is a GCE, then the
	 * latest value is retrieved from the network. If it is a feed, then the
	 * wohle array is added to the GCE.
	 * 
	 * @param href
	 *            The URI of the requested object.
	 */
	public Obj pullObj(Uri href);

	/**
	 * If the object is a transmit operation, the according GCE is looked up in the
	 * database. Else, if it is a GCE, it is looked up in the database and its
	 * group address is used to write the specified object.
	 * 
	 * @param href
	 *            The URI of the object to be pushed
	 * @param input
	 *            The object to be pushed.
	 * @param isOp
	 *            Specifies whether the object to be pushed is an operation.
	 * @throws Exception
	 *             An exception is thrown, if the object cannot be written.
	 */
	public Obj pushObj(Uri href, Obj input, boolean isOp) throws Exception;

	/**
	 * Loads objects from an XML file into the internal database.
	 * 
	 * @param file
	 *            The file containing the objects.
	 * @throws Exception
	 *             An exception is thrown, if the file could not be opened.
	 */
	public void loadObjects(File file) throws Exception;

	/**
	 * Saves objects from the internal database to an XML file.
	 * 
	 * @param file
	 *            The file containing the objects.
	 * @throws Exception
	 *             An exception is thrown, if the file could not be opened.
	 */

	public void saveObjects(File file) throws Exception;

	/**
	 * Adds the specified object to the internal database.
	 * 
	 * @param o
	 *            The object to be added.
	 */
	public void addObj(Obj o);

	/**
	 * Removes the specified object from the internal database. The object is
	 * referenced via its URI.
	 * 
	 * @param href
	 *            The URI of the object to be removed.
	 */
	public void removeObj(Uri href);

	/**
	 * Register an event listener for any object event.
	 * 
	 * @param oel
	 *            The object event listener.
	 */
	public void registerObjEventListener(IObjEventListener oel);
}
