package tuwien.auto.obix.objectbroker;

import java.io.*;
import java.util.*;
import obix.*;
import obix.io.*;
import tuwien.auto.obix.mappinglogic.*;
import tuwien.auto.obix.objects.*;

public class ObjectBroker implements IObjectBroker
{
	private IMappingLogic _ml;

	private HashMap<String, Obj> _objects;

	private MLEventListener _listener;

	/**
	 * Constructor for non-existing mapping logic
	 * 
	 */
	public ObjectBroker()
	{
		_ml = new MappingLogic();
		_objects = new HashMap<String, Obj>();
	}

	/**
	 * Constructor for existing mapping logic
	 * 
	 * @param ml
	 *            Mapping logic which shall be used.
	 */
	public ObjectBroker(IMappingLogic ml)
	{
		_ml = ml;
		_objects = new HashMap<String, Obj>();
	}

	public Obj pullObj(Uri href)
	{
		// if the path pointing to the lobby has been entered, the lobby is
		// returned
		String path = href.getPath();
		if (path.equals("/obix/"))
			return new KNXLobby();

		// else, the href references an internal object -> look it up in the
		// object database
		Obj o = _objects.get(href.getPath());

		// if the object could not be found, return an error
		if (o == null)
			return new Err("Object not found");

		// if the object is a GCE rpoint, return the latest value
		if (o.is("knx:GCE_rpoint"))
		{
			try
			{
				o = refreshGCE_rpoint((GCE_rpoint) o);
				// update value in hashtable, because the old parent object
				// reference is now broken
				removeObj(o.getChildHref("value"));
				addObj(((GCE_rpoint) o).value());
			}
			catch (Exception ex)
			{
				System.out.println(ex.getMessage());
				o = new Err("Error retrieving data from Network");
			}

		}

		// if the object's parent is a GCE rpoint, and the object is its value,
		// return the latest value
		if (o.getParent() != null)
			if (o.getName().equals("value") && (o.getParent().is("knx:GCE_rpoint")))
			{
				o = _objects.get(o.getParent().getNormalizedHref().getPath());
				try
				{
					o = refreshGCE_rpoint((GCE_rpoint) o).value();
				}
				catch (Exception ex)
				{
					System.out.println(ex.getMessage());
					o = new Err("Error retrieving data from Network");
				}
			}

		// if the object is a GCE feed, return all values
		if (o.is("knx:GCE_eventfeed"))
		{
			try
			{
				o = refreshGCE_eventfeed((GCE_eventfeed) o);
				// update feed in hashtable, because the old parent object
				// reference is now broken
				removeObj(o.getChildHref("eventfeed"));
				addObj(((GCE_eventfeed) o).eventfeed());
			}
			catch (Exception ex)
			{
				System.out.println(ex.getMessage());
				o = new Err("Error retrieving data from Network");
			}
		}

		// if the object's parent is a GCE eventfeed, and the object is its
		// feed, return the latest values
		if (o.getParent() != null)
			if (o.getName().equals("eventfeed") && (o.getParent().is("knx:GCE_eventfeed")))
			{
				o = _objects.get(o.getParent().getNormalizedHref().getPath());
				try
				{
					o = refreshGCE_eventfeed((GCE_eventfeed) o).eventfeed();
				}
				catch (Exception ex)
				{
					System.out.println(ex.getMessage());
					o = new Err("Error retrieving data from Network");
				}
			}

		return o;
	}

	/**
	 * Updates the passed GCE_eventfeed with latest data
	 * 
	 * @param gceef
	 *            GCE event feed to be updated.
	 * @return Updated event feed.
	 */
	private GCE_eventfeed refreshGCE_eventfeed(GCE_eventfeed gceef) throws Exception
	{
		// get group address
		IGroupAddress ga = (IGroupAddress) (gceef.address());
		// get datapoint type
		String dpt = gceef.eventfeed().getOf().toString();
		// retrieve datapoint event values via mapping logic
		CacheEntry[] data = null;
		data = _ml.readCache(ga, dpt);
		// add values to the feed
		Feed values = gceef.eventfeed();
		for (int i = 0; i < data.length; i++)
			values.add((Obj) (data[i].getDPValue()));
		// return the feed
		return gceef;
	}

	/**
	 * Updates the passed GCE_rpoint with latest data
	 * 
	 * @param gcerp
	 *            GCE rpoint to be updated.
	 * @return Updated GCE rpoint.
	 */
	private GCE_rpoint refreshGCE_rpoint(GCE_rpoint gcerp) throws Exception
	{
		// get group address
		IGroupAddress ga = (IGroupAddress) (gcerp.address());
		// get datapoint type
		String dpt = gcerp.value().getIs().toString();
		dpt = dpt.replace("/obix/def/", "");
		dpt = dpt.substring(0, dpt.length() - 1);
		// retrieve datapoint value via mapping logic
		IDPT data = null;
		data = _ml.readGroupAddress(ga, dpt);
		// set new value of the rpoint
		gcerp.setValue((Obj) data);
		// return the rpoint
		return gcerp;
	}

	public Obj pushObj(Uri href, Obj input, boolean isOp) throws Exception
	{
		// if the object is an operation, the according GCE is listed under the
		// href without the suffix "transmit"
		if (isOp)
			href = new Uri(href.get().replaceAll("transmit/", ""));

		// check input object
		String is = input.getIs().toString();
		if (!is.contains("/obix/def/DPT"))
			throw new Exception("Input object is not a DPT");

		// look up the object in the database
		Obj o = _objects.get(href.getPath());
		if (o == null)
			throw new Exception("Object with URI " + href.get() + " not found");
		if (!o.is("knx:GCE"))
			throw new Exception("Object with URI " + href.get() + " is not a GCE");

		// get group address
		IGroupAddress ga = (IGroupAddress) ((IGCE) o).address();

		IDPT dp = null;

		if (is.contains("DPT_B1"))
			dp = (IDPT_B1) input;

		if (is.contains("DPT_U8"))
			dp = (IDPT_U8) input;

		_ml.writeGroupAddress(ga, dp);

		return null;
	}

	public void loadObjects(File file) throws Exception
	{
		BufferedReader in = new BufferedReader(new FileReader(file));

		String objStream = new String();

		String nextline = in.readLine();
		while (!(nextline == null))
		{
			objStream += nextline;
			if ((nextline.endsWith("</obj>")) || ((nextline.contains("obj")) && (nextline.contains("/>"))))
			{
				addObj(ObixDecoder.fromString(objStream.toString()));
				objStream = new String();
			}
		}

	}

	public void saveObjects(File file) throws Exception
	{
		// TODO Auto-generated method stub

	}

	public void addObj(Obj o)
	{
		String href = o.getNormalizedHref().getPath();
		System.out.println("adding object " + href);
		_objects.put(href, o);

		// add kids
		if (o.size() > 0)
		{
			Obj[] kids = o.list();
			for (int i = 0; i < o.size(); i++)
				if (kids[i].getHref() != null)
					addObj(kids[i]);
		}
	}

	public void removeObj(Uri href)
	{
		_objects.remove(href.getPath());
	}

	public void registerObjEventListener(IObjEventListener oel)
	{
		_listener.addObjEventListener(oel);
	}
}
