/*
 * Copyright (C) 2005 Bernhard Erb

 *This program is free software; you can redistribute it and/or
 *modify it under the terms of the GNU General Public License
 *as published by the Free Software Foundation; either version 2
 *of the License, or (at your option) any later version.

 *This program is distributed in the hope that it will be useful,
 *but WITHOUT ANY WARRANTY; without even the implied warranty of
 *MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *GNU General Public License for more details.

 *You should have received a copy of the GNU General Public License
 *along with this program; if not, write to the Free Software
 *Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

package tuwien.auto.eibpoints;

import org.w3c.dom.*;

import tuwien.auto.eibxlator.*;
import tuwien.auto.eicl.struct.cemi.EIB_Address;
import tuwien.auto.eicl.util.*;

/**
 * <p>
 * This class is an abstraction of a EIB/KNX point. In this implementation the
 * group address is used for point identification. To allow further processing
 * the point is given a human readable name as well as a major and minor type
 * (DPT types). Major and minor types are used to interpret message contents
 * sent to this group address. The interpretation itself is handled by a variety
 * of classes in the eibxlator package. To enhance configuration reuse, save and
 * restore facilities are provided.
 * 
 * @author Bernhard Erb
 * @see tuwien.auto.eibpoints.PointList
 * @see tuwien.auto.eibxlator.PDUXlatorList
 * 
 */

public class Point {
    /**
     * Document element
     */
    public static final String XML_DEVICE = "point";

    private static final String XML_DEVICENAME = "point_name";

    private static final String XML_POINTTYPE_MAJOR = "pointType_Major";

    private static final String XML_POINTTYPE_MINOR = "pointType_Minor";

    private EIB_Address address;

    private String deviceName;

    private String pointType_Major;

    private String pointType_Minor;

    /**
     * This constructor is used to create a new point object. All needed
     * information are specified by respective options.
     * 
     * @param _PointName
     *            the device name
     * @param _Address
     *            the device address
     * @param _PointTypeMajor
     *            the major DPT point type key
     * @param _PointTypeMinor
     *            the minor DPT point type key
     * @throws EICLException
     *             This Exception is thrown if major or minor type can't be
     *             found.
     */
    public Point(String _PointName, EIB_Address _Address, String _PointTypeMajor,
            String _PointTypeMinor) throws EICLException {
        deviceName = _PointName;
        pointType_Major = _PointTypeMajor;
        pointType_Minor = _PointTypeMinor;
        if (PDUXlatorList.getPointTypeMajor().getTypeDescription(_PointTypeMajor) == null)
            throw new EICLException("Major Type not found");
        address = _Address;
    }

    /**
     * This constructor is used to restore a point configuration from a
     * XML_Node. If the structure is not the expected an exception is thrown.
     * 
     * @param _XMLNode
     *            The document node with label "point"
     * @throws EICLException
     *             This exception is thrown if the XML node has not the expected
     *             structure, or if the major/minor DPT type can't be found.
     */
    public Point(Node _XMLNode) throws EICLException {
        if (!_XMLNode.getNodeName().equals(XML_DEVICE))
            throw new EICLException("Node has to be a " + XML_DEVICE + " node");

        NodeList nl = _XMLNode.getChildNodes();
        if (nl.getLength() != 4)
            throw new EICLException("Node must have 4 childs" + nl.getLength());

        for (int i = 0; i < nl.getLength(); i++) {
            Element element = (Element) nl.item(i);
            if (element.getNodeName().equals(XML_DEVICENAME))
                deviceName = element.getFirstChild().getNodeValue();
            else if (element.getNodeName().equals(XML_POINTTYPE_MAJOR)) {
                pointType_Major = element.getFirstChild().getNodeValue();
            } else if (element.getNodeName().equals(XML_POINTTYPE_MINOR)) {
                pointType_Minor = element.getFirstChild().getNodeValue();
            } else if (element.getNodeName().equals(EIB_Address.XML_EIB_ADDRESS))
                address = new EIB_Address(element);
        }
    }

    /**
     * Returns the bounded EIB/KNX group addresses. In this class the returned
     * array contains only 1 element. Nevertheless this structure will be needed
     * in further versions.
     * 
     * @return the group addresses
     */
    public EIB_Address[] getDeviceAddress() {
        EIB_Address[] help = { address };
        return help;
    }

    /**
     * Sets the device address
     * 
     * @param _Address
     *            The device addresses
     */
    public void setDeviceAddress(EIB_Address _Address) {
        address = _Address;
    }

    /**
     * Returns a string composed of the device name and address string
     * 
     * @return A point description string
     */
    public String getPointString() {
        return deviceName + "   " + address.toString();
    }

    /**
     * This method is used for store the object values in XML format.
     * 
     * @param doc
     *            the XML document object
     * @return the XML document node with label "point"
     */
    public Node toXML(Document doc) {
        Element root = doc.createElement(XML_DEVICE);

        Element element = doc.createElement(XML_DEVICENAME);
        element.appendChild(doc.createTextNode(deviceName));
        root.appendChild(element);

        element = doc.createElement(XML_POINTTYPE_MAJOR);
        element.appendChild(doc.createTextNode(pointType_Major));
        root.appendChild(element);

        element = doc.createElement(XML_POINTTYPE_MINOR);
        element.appendChild(doc.createTextNode(getMinorType()));
        root.appendChild(element);

        root.appendChild(address.toXML(doc));
        return root;
    }

    /**
     * Returns the device name
     * 
     * @return the device name
     */
    public final String getDeviceName() {
        return new String(deviceName);
    }

    /**
     * Sets the device name.
     * 
     * @param _DeviceName
     *            the new device name.
     */
    public final void setDeviceName(String _DeviceName) {
        deviceName = _DeviceName;
    }

    /**
     * Returns the major type.
     * 
     * @return the major type.
     */
    public final String getMajorType() {
        return pointType_Major;
    }

    /**
     * Get the minor type.
     * 
     * @return the minor type.
     */
    public final String getMinorType() {
        return pointType_Minor;
    }

}
