/*
 * Copyright (C) 2005 Bernhard Erb

 *This program is free software; you can redistribute it and/or
 *modify it under the terms of the GNU General Public License
 *as published by the Free Software Foundation; either version 2
 *of the License, or (at your option) any later version.

 *This program is distributed in the hope that it will be useful,
 *but WITHOUT ANY WARRANTY; without even the implied warranty of
 *MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *GNU General Public License for more details.

 *You should have received a copy of the GNU General Public License
 *along with this program; if not, write to the Free Software
 *Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

package tuwien.auto.eibpoints;

import java.util.*;

import javax.xml.parsers.*;
import org.w3c.dom.*;

import tuwien.auto.eicl.struct.cemi.EIB_Address;
import tuwien.auto.eicl.util.*;

import javax.xml.transform.*;
import javax.xml.transform.dom.*;
import javax.xml.transform.stream.*;
import java.io.*;

/**
 * <p>
 * The PointList class facilitates the integration of a variety of EIB/KNX
 * points. This class provides methods for adding and removing as well as lookup
 * of EIB/KNX points by different parameters. Finally the whole configuration
 * can be saved and restored in XML format.
 * 
 * @author Bernhard Erb
 * @see tuwien.auto.eibpoints.Point
 */

public class PointList {
    private static final String XML_DEVICE = "devicelist";

    private Vector deviceList;

    /**
     * Default constructor. Use this to construct an empty list.
     */
    public PointList() {
        deviceList = new Vector();
    }

    /**
     * This constructor restores a PointList configuration, by parsing a XML
     * structure. If the XML structure is not the expected an EICLException is
     * thrown.
     * 
     * @param _XMLStream
     *            The XML file as InputStream
     * @throws EICLException
     *             The EICLException is thrown if the passed structure does not
     *             conform to the expected.
     */
    public PointList(InputStream _XMLStream) throws EICLException {
        try {
            deviceList = new Vector();
            DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
            DocumentBuilder builder = factory.newDocumentBuilder();
            Document document = builder.parse(_XMLStream);
            init(document);
        } catch (Exception ex) {
            throw new EICLException(ex.getMessage());
        }
    }

    /**
     * This constructor initializes this object by parsing a configuration file
     * in XML format.
     * 
     * @param _DeviceFile
     *            The file in which the configuration is stored
     * @throws EICLException
     *             If the structure does not correspond to the expected, or the
     *             Point constructor throws an Exception
     */
    public PointList(String _DeviceFile) throws EICLException {
        try {
            deviceList = new Vector();
            DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
            DocumentBuilder builder = factory.newDocumentBuilder();
            Document document = builder.parse(_DeviceFile);
            init(document);
        } catch (Exception ex) {
            throw new EICLException(ex.getMessage());
        }
    }

    private void init(Document document) throws EICLException {
        try {
            Element element = document.getDocumentElement();
            NodeList nl = element.getChildNodes();

            if (!element.getNodeName().equals(XML_DEVICE))
                throw new EICLException("Wrong root element: must be named xml_device");

            for (int i = 0; i < nl.getLength(); i++) {
                element = (Element) nl.item(i);
                if (element.getNodeName().equals(Point.XML_DEVICE)) {
                    Point help = new Point(element);
                    addPoint(help);
                } else
                    throw new EICLException(Point.XML_DEVICE + " node expected");
            }

        } catch (Exception ex) {
            throw new EICLException(ex.getMessage());
        }
    }

    /**
     * Add a EIB/KNX point to the actual configuration.
     * 
     * @param _Point
     *            The EIB/KNX point reference.
     * @throws EICLException
     *             If the device already exists.
     */
    public void addPoint(Point _Point) throws EICLException {
        if (getPointNr(_Point) == -1)
            deviceList.add(_Point);
        else
            throw new EICLException("Could not add device, address exists");
    }

    /**
     * Removes this EIB/KNX point.
     * 
     * @param _Point
     *            The reference to the point to be removed.
     */
    public void removePoint(Point _Point) {
        deviceList.remove(_Point);
    }

    /**
     * Get a reference to the point with the specified name.
     * 
     * @param _PointName
     *            The EIB/KNX point name.
     * @return the point Returns the corresponding point.
     * @throws EICLException
     *             The EICLException is thrown if no point with this name can be
     *             found.
     */
    public Point getPoint(String _PointName) throws EICLException {
        boolean goon = true;
        Point ret = null;
        Enumeration enumeration = deviceList.elements();
        while (enumeration.hasMoreElements() && goon) {
            Point help = (Point) enumeration.nextElement();
            if (help.getDeviceName().equals(_PointName)) {
                ret = help;
                goon = false;
            }
        }
        if (goon) {
            throw new EICLException("Devicename not found");
        }
        return ret;

    }

    /**
     * Searches the list and returns the point with the specified group address.
     * 
     * @param _Address
     *            The EIB group address.
     * @return the point The EIB/KNX point with the specified address.
     * @throws EICLException
     *             The EICLException is thrown if none can be found.
     */
    public Point getPoint(EIB_Address _Address) throws EICLException {
        boolean goon = true;
        Enumeration enumeration = deviceList.elements();
        Point ret = null;
        while (enumeration.hasMoreElements() && goon) {
            ret = (Point) enumeration.nextElement();
            EIB_Address[] addi = ret.getDeviceAddress();
            for (int i = 0; i < addi.length; i++) {
                if (_Address.toString().equals(addi[i].toString()))
                    goon = false;
            }
        }
        if (goon)
            throw new EICLException("Device not found");
        return ret;
    }

    /**
     * Returns a array with all EIB/KNX points contained in the point list. Note
     * that all parameters are passed as references.
     * 
     * @return The array containing all points
     */
    public Point[] getAllPoints() {
        Point[] point = new Point[deviceList.size()];
        Enumeration enumeration = deviceList.elements();
        int i = 0;
        while (enumeration.hasMoreElements()) {
            point[i] = (Point) enumeration.nextElement();
            i++;
        }

        return point;
    }

    /**
     * Get a description containing the name and address string from all points
     * contained in this object.
     * 
     * @return A point description of all points.
     */
    public String[] getPointDescriptions() {
        String[] ret = new String[deviceList.size()];
        for (int i = 0; i < deviceList.size(); i++) {
            ret[i] = ((Point) deviceList.get(i)).getPointString();
        }
        return ret;
    }

    private int getPointNr(Point _Point) {
        boolean goon = true;
        Enumeration enumeration = deviceList.elements();
        EIB_Address[] address = _Point.getDeviceAddress();
        int zahl = 0;
        while (enumeration.hasMoreElements() && goon) {
            Point p = (Point) enumeration.nextElement();
            EIB_Address[] addi = p.getDeviceAddress();
            for (int i = 0; i < addi.length; i++) {
                for (int f = 0; f < address.length; f++) {
                    if (addi[0].toString().equals(address[f].toString()))
                        goon = false;
                    else
                        zahl++;
                }
            }
        }
        if (goon) // we hasn't found anything
            return -1;
        return zahl;
    }

    /**
     * Stores the current configuration at the specified file location, in XML
     * format.
     * 
     * @param FileName
     *            The file location.
     * @throws EICLException
     *             If the creation was erroneous.
     */
    public void toFile(String FileName) throws EICLException {
        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        factory.setValidating(true);

        try {
            DocumentBuilder db = factory.newDocumentBuilder();
            Document doc = db.newDocument();
            TransformerFactory tFactory = TransformerFactory.newInstance();
            Transformer transform = tFactory.newTransformer();
            Element root = doc.createElement(XML_DEVICE);
            Enumeration enumeration = deviceList.elements();
            while (enumeration.hasMoreElements()) {
                root.appendChild(((Point) enumeration.nextElement()).toXML(doc));
            }
            doc.appendChild(root);
            DOMSource source = new DOMSource(doc);
            StreamResult result = new StreamResult(FileName);
            transform.transform(source, result);
        } catch (TransformerException ex) {
            throw new EICLException(ex.getMessage());
        } catch (ParserConfigurationException ex) {
            throw new EICLException(ex.getMessage());
        }

    }

    /**
     * Saves the configuration in XML file to an output stream.
     * 
     * @return The output stream containing the point list configuration in XML
     *         format.
     * @throws EICLException
     *             If the file creation was erroneous.
     */
    public OutputStream toStream() throws EICLException {
        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        factory.setValidating(true);

        try {
            DocumentBuilder db = factory.newDocumentBuilder();
            Document doc = db.newDocument();
            TransformerFactory tFactory = TransformerFactory.newInstance();
            Transformer transform = tFactory.newTransformer();
            Element root = doc.createElement(XML_DEVICE);
            Enumeration enumeration = deviceList.elements();
            while (enumeration.hasMoreElements()) {
                root.appendChild(((Point) enumeration.nextElement()).toXML(doc));
            }
            doc.appendChild(root);
            DOMSource source = new DOMSource(doc);
            OutputStream stream = new ByteArrayOutputStream();
            StreamResult result = new StreamResult(stream);
            transform.transform(source, result);
            return stream;
        } catch (TransformerException ex) {
            throw new EICLException(ex.getMessage());
        } catch (ParserConfigurationException ex) {
            throw new EICLException(ex.getMessage());
        }
    }

}
