/*
 * Copyright (C) 2005 Bernhard Erb

 *This program is free software; you can redistribute it and/or
 *modify it under the terms of the GNU General Public License
 *as published by the Free Software Foundation; either version 2
 *of the License, or (at your option) any later version.

 *This program is distributed in the hope that it will be useful,
 *but WITHOUT ANY WARRANTY; without even the implied warranty of
 *MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *GNU General Public License for more details.

 *You should have received a copy of the GNU General Public License
 *along with this program; if not, write to the Free Software
 *Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

/*
 $Log: PointPDUXlator_2ByteFloat.java,v $
 Revision 1.12  2006/01/20 20:37:24  gneug
 Some typos in DPT description entries fixed.

 */

package tuwien.auto.eibxlator;

import tuwien.auto.eicl.util.EICLException;

/**
 * <p>
 * This class implements the EIB Application Layer type EIS 5. This type is a
 * two byte floating format with the granularity of 0.01 and the max. Value of
 * 670760.
 * 
 * @author Bernhard Erb
 */

public class PointPDUXlator_2ByteFloat extends PointPDUXlator {
    /**
     * Nr. 5.001 Data Point Type Temperature; values from -273 to +670670 degrees C
     */
    public static final String[] DPT_TEMPERATURE = { "5.001", "Temperature", "-273...+670670", "°C" };

    /**
     * Nr. 5.002 Data Point Type Temperature difference; values of +/- 670760 K
     */
    public static final String[] DPT_TEMPERATURE_DIFFERENCE = { "5.002", "Temperature difference",
            "+/- 670760", "K" };

    /**
     * Nr. 5.003 Data Point Type Temperature gradient; values of +/- 670760 K/h
     */
    public static final String[] DPT_TEMPERATURE_GRADIENT = { "5.003", "Temperature gradient",
            "+/- 670760", "K/h" };

    /**
     * Nr. 5.004 Data Point Type Intensity of light; values from 0 to 670760 Lux
     */
    public static final String[] DPT_INTENSITY_OF_LIGHT = { "5.004", "Light intensity",
            "0..670760", "Lux" };

    /**
     * Nr. 5.005 Data Point Type Wind speed; values from 0 to 670760 m/s
     */
    public static final String[] DPT_WIND_SPEED = { "5.005", "Wind speed", "0..670760", "m/s" };

    /**
     * Nr. 5.006 Data Point Type Air pressure; values from 0 to 670760 Pa
     */
    public static final String[] DPT_AIR_PRESSURE = { "5.006", "Air pressure", "0..670760", "Pa" };

    /**
     * Nr. 5.010 Data Point Type Time difference1; values of +/-670760 s
     */
    public static final String[] DPT_TIME_DIFFERENCE1 = { "5.010", "Time difference 1", "+/- 670760",
            "s" };

    /**
     * Nr. 5.011 Data Point Type Time difference 2; values of +/-670760 ms
     */
    public static final String[] DPT_TIME_DIFFERENCE2 = { "5.011", "Time difference 2", "+/- 670760",
            "ms" };

    /**
     * Nr. 5.020 Data Point Type Voltage; values of +/-670760 mV
     */
    public static final String[] DPT_VOLTAGE = { "5.020", "Voltage", "+/- 670760", "mV" };

    /**
     * Nr. 5.021 Data Point Type Electrical Current; values from +/-670760 mA
     */
    public static final String[] DPT_ELECTRICAL_CURRENT = { "5.021", "Electrical current",
            "+/- 670760", "mA" };

    private static boolean sign;

    private static int mantisse = 0;

    private static byte exponent = 0;

    private static EICLMap eiclMap = null;

    /**
     * Initializes the object and sets the minor point type.
     * 
     * @param _PointType_Minor
     *            The minor point type.
     * @throws EICLException
     *             The Exception is thrown if the minor point type doesn't
     *             exist.
     */
    public PointPDUXlator_2ByteFloat(String _PointType_Minor) throws EICLException {
        initMap();
        pdu = new short[2];

        pointType_Minor = eiclMap.get(_PointType_Minor);

        if (pointType_Minor == null)
            throw new EICLException("Unknown servicetype");
    }

    /**
     * Dummy constructor only used for initialising the eiclMap. Don't use this.
     */
    public PointPDUXlator_2ByteFloat() {
        initMap();
    }

    private void initMap() {
        if (eiclMap == null) {
            eiclMap = new EICLMap();
            eiclMap.put(DPT_TEMPERATURE[0], DPT_TEMPERATURE);
            eiclMap.put(DPT_TEMPERATURE_DIFFERENCE[0], DPT_TEMPERATURE_DIFFERENCE);
            eiclMap.put(DPT_TEMPERATURE_GRADIENT[0], DPT_TEMPERATURE_GRADIENT);
            eiclMap.put(DPT_INTENSITY_OF_LIGHT[0], DPT_INTENSITY_OF_LIGHT);
            eiclMap.put(DPT_WIND_SPEED[0], DPT_WIND_SPEED);
            eiclMap.put(DPT_AIR_PRESSURE[0], DPT_AIR_PRESSURE);
            eiclMap.put(DPT_TIME_DIFFERENCE1[0], DPT_TIME_DIFFERENCE1);
            eiclMap.put(DPT_TIME_DIFFERENCE2[0], DPT_TIME_DIFFERENCE2);
            eiclMap.put(DPT_VOLTAGE[0], DPT_VOLTAGE);
            eiclMap.put(DPT_ELECTRICAL_CURRENT[0], DPT_ELECTRICAL_CURRENT);
        }
    }

    /**
     * Returns the ASDU value as float. If the ASDU has not the expected format
     * (To few bytes for example) 0 is returned.
     * 
     * @return The ASDU value turned in float value.
     */
    public float getASDUasFloat() {
        float ret = 0;
        if (pdu.length == 2) {
            if ((pdu[0] & 0x0080) == 0x0080)
                sign = false; // negative
            else
                sign = true; // positive

            exponent = (byte) ((pdu[0] & 0x0078) >> 3);
            mantisse = (((pdu[0]) & 0x0007) << 8) | ((pdu[1]) & 0x00FF);
            ret = 0;

            if (sign) {
                ret = new Double((mantisse) * 0.01 * Math.pow(2, exponent)).floatValue();
            } else {
                ret = new Double(((~(mantisse) & 0x07FF) - 1) * 0.01 * Math.pow(2, exponent)).floatValue();
            }
        }
        return ret;

    }

    /**
     * Returns the ASDU in string format with the addition of the corresponding
     * unit of measurement.
     */
    public String getASDUasString() {
        String ret = String.valueOf(getASDUasFloat()) + " " + pointType_Minor[3];
        return ret;
    }

    /**
     * Returns the APDU Byte array. This method is thought for encoding.
     */
    public byte[] getAPDUByteArray() {
        short s = (short) (servicetype << 6);
        byte[] ret = null;
        if (servicetype == A_GROUPVALUE_READ) {
            ret = new byte[2];
            ret[0] = (byte) ((s & 0xFF00) >> 8);
            ret[1] = (byte) (s & 0x00FF);
        } else if (servicetype == A_GROUPVALUE_WRITE) {
            ret = new byte[4];
            ret[0] = (byte) ((s & 0xFF00) >> 8);
            ret[1] = (byte) (s & 0x00FF);
            if (!sign)
                ret[2] |= 0x80;
            ret[2] |= ((exponent & 0x0F) << 3);
            ret[2] |= ((mantisse & 0x0700) >> 8);
            ret[3] |= ((mantisse & 0x00FF));
        }
        return ret;
    }

    /**
     * Sets the ASDU bytes from float values.
     * 
     * @param _ASDU
     *            The ASDU float value.
     * @throws EICLException
     *             If the exponent is > 15.
     */
    public void setASDUfromFloat(float _ASDU) throws EICLException {
        if (_ASDU <= 0) {
            float help = _ASDU * 100;
            exponent = 0;
            sign = true;
            while (help <= 2047) {
                exponent++;
                help /= 2;
            }
            mantisse = Math.round(help);
        } else {
            float help = _ASDU * -100;
            sign = false;
            exponent = 0;
            while (help <= 2048) {
                exponent++;
                help /= 2;
            }
            mantisse = Math.round(help);
            mantisse = ~mantisse;
            mantisse++;
        }
        if (exponent > 15)
            throw new EICLException("Exponent is > 15");

    }

    /**
     * Set the ASDU from a string value. The string is interpreted as float
     * value.
     * 
     * @param _ASDU
     *            The float value in string format
     */
    public void setASDUfromString(String _ASDU) throws EICLException {
        float f = Float.valueOf(_ASDU).floatValue();
        setASDUfromFloat(f);
    }

    /**
     * Returns the set minor type.
     * 
     * @return The minor type
     */
    public String getPointTypeMinor() {
        return pointType_Minor[0];
    }

    /**
     * Get the type specific minor types
     * 
     * @return the minor types
     */
    public EICLMap getMinorTypes() {
        return eiclMap;
    }

}
