/*
 * Copyright (C) 2005 Bernhard Erb

 *This program is free software; you can redistribute it and/or
 *modify it under the terms of the GNU General Public License
 *as published by the Free Software Foundation; either version 2
 *of the License, or (at your option) any later version.

 *This program is distributed in the hope that it will be useful,
 *but WITHOUT ANY WARRANTY; without even the implied warranty of
 *MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *GNU General Public License for more details.

 *You should have received a copy of the GNU General Public License
 *along with this program; if not, write to the Free Software
 *Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

package tuwien.auto.eibxlator;

import java.util.StringTokenizer;
import tuwien.auto.eicl.util.EICLException;

/**
 * <p>
 * This class implements the KNX DPT 3. Minor type dimming, blinds and boiler.
 * This type consists of a increase, decrease bit and a 3 bit value.
 * 
 * @author Bernahrd Erb
 */

public class PointPDUXlator_3BitControlled extends PointPDUXlator {
    /**
     * Nr. 1.007 Data Point Type Dimming Control; Values are boolean.step
     * 
     * @see PointPDUXlator_Boolean
     */
    public static final String[] DPT_Control_Dimming = { "1.007", "Dimming",
            PointPDUXlator_Boolean.DPT_STEP[2], PointPDUXlator_Boolean.DPT_STEP[3] };

    /**
     * Nr. 1.008 Data Point Type Blinds Control; Values are boolean.updown
     * 
     * @see PointPDUXlator_Boolean
     */
    public static final String[] DPT_Control_Blinds = { "1.008", "Blinds",
            PointPDUXlator_Boolean.DPT_UPDOWN[2], PointPDUXlator_Boolean.DPT_UPDOWN[3] };

    /**
     * Nr. 1.014 Data Point Type Boiler Mode; Values are boolean.inputsource
     * 
     * @see PointPDUXlator_Boolean
     */
    public static final String[] DPT_Mode_Boiler = { "1.014", "Boiler",
            PointPDUXlator_Boolean.DPT_INPUTSOURCE[2], PointPDUXlator_Boolean.DPT_INPUTSOURCE[3] };

    private boolean increase;

    private short value;

    private static EICLMap eiclMap = null;

    /**
     * Inits the object with the given minor type
     * 
     * @param _PointType_Minor_Key
     *            the requested minor type key
     * @throws EICLException
     *             if the key doesn't exist
     */
    public PointPDUXlator_3BitControlled(String _PointType_Minor_Key) throws EICLException {
        initMap();
        pdu = new short[1];

        pointType_Minor = eiclMap.get(_PointType_Minor_Key);

        if (pointType_Minor == null)
            throw new EICLException("Unknown service type");
    }

    /**
     * Dummy constructor only used for initialising the eiclMap. Don't use this.
     */
    public PointPDUXlator_3BitControlled() {
        initMap();
    }

    private void initMap() {
        if (eiclMap == null) {
            eiclMap = new EICLMap();
            eiclMap.put(DPT_Control_Dimming[0], DPT_Control_Dimming);
            eiclMap.put(DPT_Control_Blinds[0], DPT_Control_Blinds);
            eiclMap.put(DPT_Mode_Boiler[0], DPT_Mode_Boiler);
        }
    }

    /**
     * Returns the encapsulated ASDU value scaled in percent as float.
     * 
     * @return The ASDU as float value.
     */
    public float getASDUValueFieldScaled() {
        float value = (float) (pdu[0] & 0x0007) / 7 * 100;
        return value;
    }

    /**
     * Return the ASDU value unscaled as short.
     * 
     * @return The ASDU value.
     */
    public short getASDUValueFieldUnscaled() {
        short value = (short) (pdu[0] & 0x0007);
        return value;
    }

    /**
     * Retrieves the increase decrease boolean value
     * 
     * @return True if the value is decrease. False if the message is an
     *         increase value.
     */
    public boolean getASDUControlBit() {
        boolean ret = true;
        if ((pdu[0] & 0x0008) == 0)
            ret = false;
        return ret;
    }

    /**
     * Sets the ASDU control bit to the specified direction. false is
     * interpreted as negative (decrease), whereas true is interpreted as
     * positive value(increase).
     * 
     * @param _Direction
     *            The ASDU direction to be set.
     */
    public void setASDUControlBit(boolean _Direction) {
        increase = _Direction;
    }

    /**
     * Returns a String composed by the control bit value, the step value and
     * the to the minor type corresponding unit of measurement.
     * 
     * @return The string interpretation of the ASDU.
     */
    public String getASDUasString() {
        StringBuffer sb = new StringBuffer();
        boolean decrease = !getASDUControlBit();
        if ((pdu[0] & 0x0007) == 0)
            sb.append("step break");
        else {
            if (pointType_Minor[0].equals(DPT_Control_Dimming[0])) {
                if (!decrease)
                    sb.append(PointPDUXlator_Boolean.DPT_STEP[3]);
                else
                    sb.append(PointPDUXlator_Boolean.DPT_STEP[2]);
                sb.append(" " + getASDUValueFieldScaled() + "%");
            } else if (pointType_Minor[0].equals(DPT_Control_Blinds[0])) {
                if (!decrease)
                    sb.append(PointPDUXlator_Boolean.DPT_UPDOWN[3]);
                else
                    sb.append(PointPDUXlator_Boolean.DPT_UPDOWN[2]);
                sb.append(" " + getASDUValueFieldUnscaled() + "%");
            } else if (pointType_Minor[0].equals(DPT_Mode_Boiler[0])) {
                if (!decrease)
                    sb.append(PointPDUXlator_Boolean.DPT_INPUTSOURCE[3]);
                else
                    sb.append(PointPDUXlator_Boolean.DPT_INPUTSOURCE[2]);
                sb.append(" " + getASDUValueFieldUnscaled() + "%");
            }

        }
        return sb.toString();
    }

    /**
     * Converts the input values in the requested APDU/ASDU format. The result
     * is the complete Application Layer encoding.
     * 
     * @return the asdu byte[]
     */
    public byte[] getAPDUByteArray() {
        pdu = new short[1];
        pdu[0] = (short) (servicetype << 6);
        if (servicetype == A_GROUPVALUE_WRITE) {
            if (increase)
                pdu[0] |= 0x0008;
            else
                pdu[0] &= 0xFFF7;
            pdu[0] |= (value & 0x007);
        }
        byte[] ret = { (byte) ((pdu[0] & 0xFF00) >> 8), (byte) (pdu[0] & 0x00FF) };
        return ret;
    }

    /**
     * Sets the ASDU value from a unsigned unscaled short value. Note that 000
     * corresponds to a step break.
     * 
     * @param stepValue
     *            The ASDU value.
     * @throws EICLException
     *             If the step value is outside 0..7.
     */
    public void setASDUValueFieldUnsigned(short stepValue) throws EICLException {
        if (stepValue < 0 || stepValue > 7)
            throw new EICLException("Value must be inside [0..7]");
        value = stepValue;
    }

    /**
     * Sets the ASDU value form a scaled float value. Note that 000 corresponds
     * to a step break.
     * 
     * @param stepValue
     *            The ASDU value.
     * @throws EICLException
     *             If the step value is outside 0..7.
     */
    public void setASDUValueFieldScaled(float stepValue) throws EICLException {
        short help = new Float(stepValue * 7 / 100).shortValue();
        if (help < 0 || help > 7)
            throw new EICLException("Value must be inside [0..7]");
        value = help;
    }

    /**
     * Sets the ASDU from a String. The string must be composed of the
     * corresponding minor type control flag value separated by a blank from the
     * value itself. Note that all values are interpreted unsigned short.
     */
    public void setASDUfromString(String _ASDU) throws EICLException {
        StringTokenizer token = new StringTokenizer(_ASDU);
        String pdu = token.nextToken();
        if (pdu.equals(PointPDUXlator_Boolean.DPT_INPUTSOURCE[3])
                || pdu.equals(PointPDUXlator_Boolean.DPT_UPDOWN[3])
                || pdu.equals(PointPDUXlator_Boolean.DPT_STEP[3]))
            increase = true;
        else if (pdu.equals(PointPDUXlator_Boolean.DPT_INPUTSOURCE[2])
                || pdu.equals(PointPDUXlator_Boolean.DPT_UPDOWN[2])
                || pdu.equals(PointPDUXlator_Boolean.DPT_STEP[2]))
            increase = false;
        else
            throw new EICLException("Pdu string not found");
        setASDUValueFieldUnsigned(Short.valueOf(token.nextToken()).shortValue());
    }

    /**
     * Returns the set minor type.
     * 
     * @return the minor type
     */
    public String getPointTypeMinor() {
        return pointType_Minor[0];
    }

    /**
     * Get the type specific minor types
     * 
     * @return the minor types
     */
    public EICLMap getMinorTypes() {
        return eiclMap;
    }

}
