/*
 * Copyright (C) 2005 Bernhard Erb

 *This program is free software; you can redistribute it and/or
 *modify it under the terms of the GNU General Public License
 *as published by the Free Software Foundation; either version 2
 *of the License, or (at your option) any later version.

 *This program is distributed in the hope that it will be useful,
 *but WITHOUT ANY WARRANTY; without even the implied warranty of
 *MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *GNU General Public License for more details.

 *You should have received a copy of the GNU General Public License
 *along with this program; if not, write to the Free Software
 *Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

/*
$Log: PointPDUXlator_8BitUnsigned.java,v $
Revision 1.10  2006/01/12 21:37:33  gneug
DPT 5.004 removed, since it was confusing (0..255%) and not for general use

 */

package tuwien.auto.eibxlator;

import tuwien.auto.eicl.util.EICLException;

/**
 * <p>
 * This class implements the KNX DPT 6 Application Layer type. The values are in
 * the range of 0-255.
 * 
 * @author Bernhard Erb
 */

public class PointPDUXlator_8BitUnsigned extends PointPDUXlator {
    /**
     * Nr. 5.001 Data Point Type Scaling; values from 0 to 100 %
     */
    public static final String[] DPT_SCALING = { "5.001", "DPT_Scaling", "[0..100]", "%" };

    /**
     * Nr. 5.003 Data Point Type Angle; values form 0 to 360 %
     */
    public static final String[] DPT_ANGLE = { "5.003", "DPT_Angle", "[0..360]", "°" };

    /*
     * Nr. 5.004: values from 0 to 255 % (sic), for use within functional block only
     */

    /**
     * Nr. 5.010 Data Point Type value 1 ucount; values from 0 to 255 counter
     * pulses
     */
    public static final String[] DPT_VALUE_1_UCOUNT = { "5.010", "DPT_Value_1_Ucount", "[0..255]",
            "counter pulses" };

    private String[] pointType_Minor;

    private static EICLMap eiclMap = null;

    /**
     * Inits the object with the requested minor type.
     * 
     * @param _PointType_Minor_Key
     *            the minor type key.
     * @throws EICLException
     *             If the minor type key doesn't exist.
     */
    public PointPDUXlator_8BitUnsigned(String _PointType_Minor_Key) throws EICLException {
        initMap();
        pdu = new short[1];

        pointType_Minor = eiclMap.get(_PointType_Minor_Key);
        if (pointType_Minor == null)
            throw new EICLException("Minor Type not found");
    }

    /**
     * Dummy constructor only used for initialising the eiclMap. Don't use this
     */
    public PointPDUXlator_8BitUnsigned() {
        initMap();
    }

    private void initMap() {
        if (eiclMap == null) {
            eiclMap = new EICLMap();
            eiclMap.put(DPT_SCALING[0], DPT_SCALING);
            eiclMap.put(DPT_ANGLE[0], DPT_ANGLE);
            eiclMap.put(DPT_VALUE_1_UCOUNT[0], DPT_VALUE_1_UCOUNT);
        }
    }

    /**
     * Sets the ASDU byte from a unsigned unscaled short.
     * 
     * @param _ASDU
     *            The ASDU byte.
     */
    public void setASDUfromShortUnscaled(short _ASDU) {
        pdu[0] = (byte) _ASDU;
    }

    /**
     * Sets the ASDU byte form a float. Note that DPT_SCALING and DPT_ANGLE
     * values are interpreted scaled.
     * 
     * @param _ASDU
     *            The ASDU short value.
     * @throws EICLException
     *             If the conversion didn't succeed.
     */
    public void setASDUfromShort(short _ASDU) throws EICLException {
        try {
            if (pointType_Minor.equals(DPT_SCALING)) {
                pdu[0] = (short) (new Short(_ASDU).floatValue() * 255 / 100);
            } else if (pointType_Minor.equals(DPT_ANGLE)) {
                pdu[0] = (short) (new Short(_ASDU).floatValue() * 255 / 360);
            } else if (pointType_Minor.equals(DPT_VALUE_1_UCOUNT)) {
                pdu[0] = (short) (new Short(_ASDU).floatValue());
            }
        } catch (Exception ex) {
            throw new EICLException(ex.getMessage());
        }
    }

    /**
     * Sets the ASDU byte by parsing a string. The String is interpreted as
     * short value. See setASDUfromShort() method for details.
     */
    public void setASDUfromString(String _ASDU) throws EICLException {
        setASDUfromShort(Short.valueOf(_ASDU).shortValue());
    }

    /**
     * Returns the ASDU byte as short value. Note that DPT_SCALING and DPT_ANGLE
     * are scaled.
     * 
     * @return the ASDU as short.
     */
    public short getASDUasShort() {
        short help = 0;
        if (pointType_Minor.equals(DPT_SCALING)) {
            help = (short) ((float) (pdu[0] & 0x00FF) * 100 / 255);
        } else if (pointType_Minor.equals(DPT_ANGLE)) {
            help = (short) ((float) (pdu[0] & 0x00FF) * 360 / 255);
        } else if (pointType_Minor.equals(DPT_VALUE_1_UCOUNT)) {
            help = (short) ((pdu[0] & 0x00FF));
        }

        return help;
    }

    /**
     * The ASDU as short unscaled value.
     * 
     * @return The ASDU as short value.
     */
    public short getASDUasShortUnscaled() {
        return (short) (pdu[0] & 0x00FF);
    }

    /**
     * Return a string interpretation of the ASDU byte. The string is composed
     * of the short value and the corresponding unit of measurement.
     */
    public String getASDUasString() {
        String ret = String.valueOf(getASDUasShort()) + " " + pointType_Minor[3];
        return ret;
    }

    /**
     * Converts the input values in the requested APDU/ASDU format. The result
     * is the complete Application Layer encoding.
     * 
     * @return the ASDU byte[]
     */
    public byte[] getAPDUByteArray() {
        short s = (short) (servicetype << 6);
        byte[] ret = null;
        if (servicetype == A_GROUPVALUE_READ) {
            ret = new byte[2];
            ret[0] = (byte) ((s & 0xFF00) >> 8);
            ret[1] = (byte) (s & 0x00FF);
        }
        if (servicetype == A_GROUPVALUE_WRITE) {
            ret = new byte[3];
            ret[0] = (byte) ((s & 0xFF00) >> 8);
            ret[1] = (byte) (s & 0x00FF);
            ret[2] = (byte) pdu[0];
        }
        return ret;
    }

    /**
     * Returns the set minor type.
     * 
     * @return the minor type
     */
    public String getPointTypeMinor() {
        return pointType_Minor[0];
    }

    /**
     * Get the type specific minor types
     * 
     * @return the minor types
     */
    public EICLMap getMinorTypes() {
        return eiclMap;
    }

}
