/*
 This program is free software; you can redistribute it and/or modify it under the terms of 
 the GNU General Public License as published by the Free Software Foundation; either 
 version 2 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
 See the GNU General Public License for more details.

 You should have received a copy of the GNU General Public License along with this program; 
 if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 Boston, MA  02110-1301, USA.
 */

/*
 Copyright (C) 2005 Bernhard Erb
 Copyright (C) 2006 Georg Neugschwandtner
 */

/*
 $Log: PointPDUXlator_String.java,v $
 Revision 1.1  2006/01/11 19:05:10  gneug
 PointPDUXlator_String added

 */

package tuwien.auto.eibxlator;

import java.nio.charset.Charset;
import java.nio.ByteBuffer;
import tuwien.auto.eicl.util.EICLException;

/**
 * <p>
 * This class implements KNX DPT 16 (String) with minor types 000 (ASCII string)
 * and 001 (ISO 8859-1 encoded string).
 */

public class PointPDUXlator_String extends PointPDUXlator {
    /**
     * Nr. 16.000 Datapoint Type DPT_String_ASCII
     */
    public static final String[] DPT_String_ASCII = { "16.000", "ASCII" };

    /**
     * Nr. 16.001 Datapoint Type DPT_String_8859_1
     */
    public static final String[] DPT_String_8859_1 = { "16.001", "ISO 8859-1" };

    private String content;

    private Charset charset;

    private static EICLMap eiclMap = null;

    /**
     * Inits the object with the given minor type
     * 
     * @param _PointType_Minor_Key
     *            the requested minor type key ("16.000", "16.001")
     * @throws EICLException
     *             if the key is invalid
     */
    public PointPDUXlator_String(String _PointType_Minor_Key) throws EICLException {
        initMap();
        pdu = new short[1];

        pointType_Minor = eiclMap.get(_PointType_Minor_Key);

        if (pointType_Minor == null) {
            throw new EICLException("Unknown minor type");
        }

        // 8859-1 and US-ASCII must be supported by every Java implementation
        if (_PointType_Minor_Key.equals("16.001")) {
            charset = Charset.forName("ISO-8859-1");
        } else {
            charset = Charset.forName("US-ASCII");
        }

    }

    /**
     * Dummy constructor only used for initialising the eiclMap. Don't use this.
     */
    public PointPDUXlator_String() {
        initMap();
    }

    private void initMap() {
        if (eiclMap == null) {
            eiclMap = new EICLMap();
            eiclMap.put(DPT_String_ASCII[0], DPT_String_ASCII);
            eiclMap.put(DPT_String_8859_1[0], DPT_String_8859_1);
        }
    }

    /**
     * Returns the String represented by the ASDU.
     * 
     * @return The string interpretation of the ASDU.
     */
    public String getASDUasString() {

        int strlen;

        // find first NULL character in ASDU
        // (pdu actually contains the ASDU)
        for (strlen = 0; (strlen < 14) && (pdu[strlen] != 0); strlen++)
            ;

        ByteBuffer bb = ByteBuffer.allocate(strlen);
        
        for (int i = 0; (i < strlen); i++) {
            bb.put((byte) pdu[i]);
        }

        bb.rewind();

        return charset.decode(bb).toString();
    }

    /**
     * Converts the input values to the KNX/EIB network format.
     * 
     * The string defined via setASDUfromString() is converted to US-ASCII or
     * ISO-8859-1 as defined by the chosen minor type. Unmappable characters are
     * silently replaced, typically with question marks.
     * 
     * @return The TPDU, which is always 16 bytes in size for this DPT (1 octet
     *         TPCI/APCI, 1 octet APCI, 14 octets data).
     */
    public byte[] getAPDUByteArray() {
        // this method is a misnomer for all Xlators, since it also returns the
        // TPCI

        // TPCI is always 0 (T_DATA_GROUP_REQ)

        // this code should somehow be refactored into PointPDUXlator for use by
        // all its subclasses

        short s = (short) (servicetype << 6);
        byte[] ret = null;
        if (servicetype == A_GROUPVALUE_READ) {
            ret = new byte[2];
            ret[0] = (byte) ((s & 0xFF00) >> 8);
            ret[1] = (byte) (s & 0x00FF);
        } else if (servicetype == A_GROUPVALUE_WRITE) {
            ret = new byte[16];
            ret[0] = (byte) ((s & 0xFF00) >> 8);
            ret[1] = (byte) (s & 0x00FF);

            byte[] ba = charset.encode(content).array();

            for (int i = 0; i < 14; i++) {
                if (i < ba.length) {
                    ret[2 + i] = ba[i];
                } else {
                    ret[2 + i] = 0;
                    // unused characters have to be set to 0,
                    // ASDU length is always 14
                }
            }
        }
        return ret;

    }

    /**
     * Sets the ASDU from a String.
     * 
     * @param content
     *            The string to be converted.
     * @throws EICLException
     *             if the maximum length is exceeded.
     */
    public void setASDUfromString(String content) throws EICLException {

        if (content.length() > 14) {
            throw new EICLException("Maximum string length is 14");
        }

        this.content = content;
    }

    /**
     * Returns the set minor type.
     * 
     * @return the minor type
     */
    public String getPointTypeMinor() {
        return pointType_Minor[0];
    }

    /**
     * Get the type specific minor types
     * 
     * @return the minor types
     */
    public EICLMap getMinorTypes() {
        return eiclMap;
    }

}
