/*
 * Copyright (C) 2005 Bernhard Erb

 *This program is free software; you can redistribute it and/or
 *modify it under the terms of the GNU General Public License
 *as published by the Free Software Foundation; either version 2
 *of the License, or (at your option) any later version.

 *This program is distributed in the hope that it will be useful,
 *but WITHOUT ANY WARRANTY; without even the implied warranty of
 *MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *GNU General Public License for more details.

 *You should have received a copy of the GNU General Public License
 *along with this program; if not, write to the Free Software
 *Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

package tuwien.auto.eicl;

import java.net.*;
import java.util.*;

import java.io.*;

import tuwien.auto.eicl.struct.eibnetip.Description_Request;
import tuwien.auto.eicl.struct.eibnetip.Description_Response;
import tuwien.auto.eicl.struct.eibnetip.EIBnetIPPacket;
import tuwien.auto.eicl.struct.eibnetip.Search_Request;
import tuwien.auto.eicl.struct.eibnetip.Search_Response;
import tuwien.auto.eicl.struct.eibnetip.util.EIBNETIP_Constants;
import tuwien.auto.eicl.util.*;

/**
 * <p>
 * The Discoverer class handles EIBnet/IP search and description requests. To be
 * able to accept search and description responses, this class maintains its own
 * socket and runs as an independent thread. This is, however, not of concern
 * for the client application since thread instantiation is done implicitly in
 * the constructor.
 * <p>
 * 
 * @author Bernhard Erb
 * @see tuwien.auto.eicl.CEMI_Connection
 * @see tuwien.auto.eicl.Heartbeat
 */

public class Discoverer implements Runnable {

    private MulticastSocket socket;

    private Thread receiveThread;

    private Vector searchResponse = new Vector();

    private InetSocketAddress descriptionFrom; // the HostAddress of the
                                                // desired Description_Address

    private Description_Response description = null; // holds the last
                                                        // received
                                                        // Description_Response;

    /**
     * Opens a new socket with a random port number and establishes the new
     * instance as a receiver thread. If something goes wrong an Exception is
     * thrown.
     */

    public Discoverer() {
        socket = Util.openNextPort();
        receiveThread = new Thread(this);
        receiveThread.start();
    }

    /**
     * Opens a socket on the given port and establishes the new instance as a
     * receiver thread.
     * 
     * @param _LocalPort
     *            the socket port
     * @throws EICLException
     */
    public Discoverer(int _LocalPort) throws EICLException {
        try {
            socket = new MulticastSocket(_LocalPort);
            receiveThread = new Thread(this);
            receiveThread.start();
        } catch (Exception ex) {
            throw new EICLException(ex.getMessage());
        }
    }

    /**
     * Sends a description request message and blocks until either the answer
     * has been received or the timeout is reached. If the timeout is reached an
     * Exception is thrown.
     * 
     * @param _SocketAddress
     *            The server socket address
     * @param _Timeout
     *            Timeout (in seconds)
     * @return The description response message
     * @throws EICLException
     *             If the timeout was reached or something went wrong
     */
    public Description_Response getDescription(InetSocketAddress _SocketAddress, int _Timeout)
            throws EICLException {
        Description_Request request = new Description_Request(socket.getLocalPort());
        descriptionFrom = _SocketAddress;
        description = null;
        try {
            DatagramPacket packet = new DatagramPacket(request.toByteArray(),
                    request.toByteArray().length, _SocketAddress);
            socket.send(packet);
        } catch (Exception ex) {
            throw new EICLException(ex.getMessage());
        }

        synchronized (this) {
            try {
                wait(_Timeout * 1000);
            } catch (InterruptedException ex) {
                // do nothing
            }
        }

        if (description != null)
            return description;

        throw new EICLException("Description not received");

    }

    /**
     * Sends a search request message (multicast). The results can be accessed
     * with the getSearchResponses() method. Note that a new search request
     * causes getSearchResponses to forget previously received responses.
     * 
     * @throws EICLException
     */
    public void sendSearchRequest() throws EICLException {
        searchResponse = new Vector();
        Search_Request request = new Search_Request(socket.getLocalPort());
        try {
            InetAddress address = InetAddress.getByName(EIBNETIP_Constants.EIBNETIP_SERVER_MULTICAST);
            DatagramPacket packet = new DatagramPacket(request.toByteArray(),
                    request.toByteArray().length, address, EIBNETIP_Constants.EIBNETIP_PORT_NUMBER);
            socket.send(packet);
        } catch (IOException ex) {
            /*
             * We assume that the exception occurred because there is no
             * external network interface. Therefore we try the local loopback
             * instead.
             */
            try {
                InetAddress address = InetAddress.getByName("127.0.0.1");
                DatagramPacket packet = new DatagramPacket(request.toByteArray(),
                        request.toByteArray().length, address,
                        EIBNETIP_Constants.EIBNETIP_PORT_NUMBER);
                socket.send(packet);
            } catch (IOException exc) {
                throw new EICLException("could not send packet");
            }
        }

    }

    /**
     * Returns the socket addresses of the hosts that have answered the search
     * request. You can invoke this method multiple times to check for new
     * responses, only a new search request will erase the list of responses.
     * 
     * @return fount hosts
     */
    public InetSocketAddress[] getSearchResponses() {
        InetSocketAddress[] isa = new InetSocketAddress[searchResponse.size()];
        synchronized (searchResponse) {
            isa = (InetSocketAddress[]) searchResponse.toArray(isa);
        }
        return isa;
    }

    /**
     * The run method is responsible for receiving search and description
     * messages. It therefore blocks on the socket until a message arrives. It's
     * not necessary to call it on your own; this is done by the constructor.
     */
    public void run() {
        byte[] buffer = new byte[1024];
        DatagramPacket datapacket = new DatagramPacket(buffer, buffer.length);

        while (true) {
            try {
                socket.receive(datapacket);
                EIBnetIPPacket packet = new EIBnetIPPacket(datapacket.getData(),
                        datapacket.getLength());
                switch (packet.getServiceType()) {
                case EIBNETIP_Constants.SEARCH_RESPONSE: {
                    Search_Response response = new Search_Response(packet.getBody());
                    synchronized (searchResponse) {
                        searchResponse.add(response.getDiscoveredEndpoint().getSocketAddress());
                    }
                    break;
                }
                case EIBNETIP_Constants.DESCRIPTION_RESPONSE: {
                    if (datapacket.getSocketAddress().equals(descriptionFrom)) {
                        description = new Description_Response(packet.getBody());
                        synchronized (this) {
                            // void MI_Description_actionPerformed(ActionEvent
                            // e) {

                            notify();
                        }
                    }
                    break;
                }
                default: {
                    break;
                }

                }
            } catch (Exception ex) {
                ex.printStackTrace();
            }
        }

    }

}
