/*
 Copyright (C) 2005-2006 Bernhard Erb and contributors

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2 
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 
 */

/* 
 $Log: CEMI_L_DATA.java,v $
 Revision 1.15  2006/01/18 14:20:32  gneug
 Clearing of repeat flag in setRepeatFlag() fixed (wrong mask). Reported by Andres Gehles.

 Revision 1.14  2006/01/16 09:33:57  gneug
 Destination address type bit setting in (short, EIB_Address, EIB_Address, byte[]) constructor fixed. Reported by Martin Rau.

 Revision 1.13  2006/01/11 19:04:14  gneug
 Length limitation in (short, EIB_Address, EIB_Address, byte[]) constructor fixed (16 instead of 15 bytes for TPDU data)

 Revision 1.12  2005/12/28 09:36:24  gneug
 Documentation added, isRepeatFrame() -> isRepeatedFrame()

 Revision 1.11  2005/12/27 19:07:20  gneug
 Handling of routing counter fixed: Frames now have a default hop count of 6, shift in getHopCount() fixed, setHopCount() added

 Revision 1.10  2005/12/14 22:55:19  gneug
 Eclipse code formatter applied to all source files

 Revision 1.9  2005/12/02 17:30:12  gneug
 Type of message code parameter in constructor aligned with type of static message code fields (short)

 */

package tuwien.auto.eicl.struct.cemi;

import java.io.*;
import tuwien.auto.eicl.util.*;

/**
 * <p>
 * This class encapsulates a CEMI_L_DATA message. It has all necessary get and
 * set methods for creating or extracting a CEMI_L_DATA message. Note that this
 * class handles only standard frame formats, with application layer data up to
 * 16 bytes. Frames with additional information are not supported. CEMI_L_DATA
 * messages are typically sent and received through tunnelling connections.
 * 
 * @author Bernhard Erb, Georg Neugschwandtner
 */

public class CEMI_L_DATA implements CEMI {

    // priority codes

    /**
     * cEMI Message with system priority
     */
    public static final byte PRIORITY_SYSTEM = 0x00;

    /**
     * cEMI Message with urgent priority
     */
    public static final byte PRIORITY_URGENT = (byte) 0x80;

    /**
     * cEMI Message with normal priority
     */
    public static final byte PRIORITY_NORMAL = (byte) 0x40;

    /**
     * cEMI Message with low priority
     */
    public static final byte PRIORITY_LOW = (byte) 0xC0;

    // message codes

    /**
     * cEMI L_DATA_Request
     */
    public static final short MC_L_DATAREQ = 0x11;

    /**
     * cEMI L_DATA_Confirmation
     */
    public static final short MC_L_DATACON = 0x2E;

    /**
     * cEMI L_DATA_Indication
     */
    public static final short MC_L_DATAIND = 0x29;

    // internal fields

    // cEMI L_Data basic message structure (1 octet each unless otherwise
    // noted; AN033 2.5.3.3 and AN066 2.2.1.8 ff):
    // 
    // Message Code
    // Additional Info Length
    // Additional Information ([Additional Info Length] octets)
    // Control field 1
    // Control field 2
    // Source Address High
    // Source Address Low
    // Destination Address High
    // Destination Address Low
    // Information-length
    // TPCI/APCI/data ([Information-length] + 1 octets)

    // Control field 1, MSB to LSB:
    // frame type (0 .. extended frame, 1 .. standard frame)
    // reserved (0)
    // repeat flag (0 .. repeated frame and server did not receive
    // // original transmission, 0 .. not repeated;
    // // considered as request for automatic retransmission when sending)
    // system broadcast flag
    // // (only applicable on open media, 0 .. system broadcast, 1 .. broadcast)
    // priority (2) (11 .. lowest priority)
    // acknowledge request (1 .. L2-ACK requested, 0 .. don't care;
    // // on TP, for L_DATA.req only)
    // confirm (for L_DATA.con only, 1 .. error, 0 .. no error)

    // Control field 2, MSB to LSB: destination address type, hop count (3),
    // extended frame format (4)

    private short messagecode;

    private short addilength;

    private short ctrl1;

    private short ctrl2;

    private EIB_Address saddr;

    private EIB_Address daddr;

    private short apci_length;

    private byte[] data;

    /**
     * Creates a new CEMI_L_DATA frame. Only standard frames are accepted.
     * 
     * @param _Messagecode
     *            CEMI_L_DATA message code
     * @param _Source
     *            individual Source address
     * @param _Dest
     *            destination address
     * @param _Data
     *            message body (starting with transport layer control
     *            information)
     * @see EIB_Address
     * @exception EICLException
     *                if the TPDU byte array is larger than 16 bytes.
     */
    public CEMI_L_DATA(short _Messagecode, EIB_Address _Source, EIB_Address _Dest, byte[] _Data)
            throws EICLException {
        if ((_Messagecode != MC_L_DATAREQ) && (_Messagecode != MC_L_DATACON)
                && (_Messagecode != MC_L_DATAIND))
            throw new EICLException("Unknown message code");
        messagecode = _Messagecode;
        addilength = 0;
        ctrl1 = 0;
        ctrl2 = 0x60; // default: routing counter = 6
        setPriority(PRIORITY_NORMAL);
        if (_Dest.isGroupAddress())
            ctrl2 |= 0x0080;
        else
            ctrl2 &= 0xFF7F;
        setFrameFormatStandard();
        saddr = _Source;
        if (saddr.isGroupAddress())
            throw new EICLException("Source address can't be of type group address");
        daddr = _Dest;
        data = _Data;
        apci_length = (short) data.length;

        if (data.length > 16)
            throw new EICLException(
                    "Maximum allowable TPDU length is 16 octets in this implementation");
    }

    /**
     * Creates and parses a new CEMI_L_DATA frame out of a byte array. Only
     * standard frames are supported
     * 
     * @param _CEMI
     *            UDP message body
     * @exception EICLException
     *                If the provided frame is not a standard frame an exception
     *                is thrown.
     */
    public CEMI_L_DATA(byte[] _CEMI) throws EICLException {
        ByteArrayInputStream bais = new ByteArrayInputStream(_CEMI);
        messagecode = (short) bais.read();
        addilength = (short) bais.read();
        if (addilength != 0) {
            throw new EICLException("Only cEMI packets with additional info length 0 are supported");
        }
        ctrl1 = (short) bais.read();
        ctrl2 = (short) bais.read();
        byte[] help = new byte[2];
        bais.read(help, 0, 2);
        saddr = new EIB_Address(help, false);
        bais.read(help, 0, 2);

        if ((ctrl2 & 0x0080) == 0x0080)
            daddr = new EIB_Address(help, true);
        else
            daddr = new EIB_Address(help, false);

        apci_length = (short) bais.read();
        data = new byte[apci_length + 1];
        for (int i = 0; i < apci_length + 1; i++) {
            data[i] = (byte) bais.read();
        }

        if (isExtendedFrame())
            throw new EICLException("Only standard frames are supported");

    }

    /**
     * Returns the byte representation of this message. If an IOException
     * occurred the EICLExcepiton is thrown.
     * 
     * @see tuwien.auto.eicl.struct.cemi.CEMI#toByteArray()
     * @exception EICLException
     *                If an IOException occurs.
     */

    public byte[] toByteArray() throws EICLException {
        ByteArrayOutputStream baos = new ByteArrayOutputStream();

        try {
            baos.write((byte) messagecode);
            baos.write((byte) addilength);
            baos.write((byte) ctrl1);
            baos.write((byte) ctrl2);
            baos.write(saddr.toByteArray());
            baos.write(daddr.toByteArray());
            baos.write((byte) (apci_length - 1));
            baos.write(data);
        } catch (IOException ex) {
            throw new EICLException(ex.getMessage());
        }

        return baos.toByteArray();
    }

    /**
     * Returns the CEMI_L_DATA message code. See constants.
     * 
     * @return CEMI_L_DATA message code.
     */

    public short getMessageCode() {
        return messagecode;
    }

    /**
     * Get the CEMI_L_DATA source address.
     * 
     * @return Returns the source address
     */
    public EIB_Address getSourceAddress() {
        return saddr;
    }

    /**
     * Get the CEMI_L_DATA destination address.
     * 
     * @return Returns the destination address.
     */
    public EIB_Address getDestinationAddress() {
        return daddr;
    }

    /**
     * Get the CEMI_L_DATA application layer data (Message body).
     * 
     * @return application layer data
     */
    public byte[] getData() {
        return data;
    }

    /**
     * Checks whether the frame is an extended or normal frame. In the current
     * implementation this method always returns false.
     * 
     * @return Is the current frame in extended frame format?
     */
    public boolean isExtendedFrame() {
        if ((ctrl1 & 0x0080) == 0)
            return true;
        return false;
    }

    /**
     * Returns the state of the repeat flag. Especially relevant for L_Data
     * indications. Note that a cEMI server should not pass incoming repeated
     * frames to clients when it has already received (and passed on) the
     * corresponding original, non-repeated frame.
     * 
     * @return True if the repeat flag is 0 (zero), false otherwise.
     */
    public boolean isRepeatedFrame() {
        return ((ctrl1 & 0x20) == 0);
    }

    /**
     * Sets the repeat flag. The repeat flag is considered a request for
     * automatic retransmission on L_Data requests.
     * 
     * @param _Repeat
     *            True to set the repeat flag 0 (repeated on medium/do not
     *            repeat on error). False to set the repeat flag 1 (not repeated
     *            on medium/don't care).
     */
    public void setRepeatFlag(boolean _Repeat) {
        if (_Repeat)
            ctrl1 &= 0xDF;
        else
            ctrl1 |= 0x20;
    }

    /**
     * Get the message priority
     * 
     * @return Returns the message priority
     */
    public byte getPriority() {
        return (byte) ((ctrl1 & 0x000C) >> 8);
    }

    /**
     * Set the message priority. See constants.
     * 
     * @param _Priority
     *            Sets the message priority.
     */
    public void setPriority(byte _Priority) {
        ctrl1 |= _Priority;
    }

    /**
     * Checks whether the message was acknowledged by an endpoint.
     * 
     * @return the acknowledge bit
     */
    public boolean getAcknowledgeRequested() {
        if ((ctrl1 & 0x0002) == 0)
            return false;
        return true;
    }

    /**
     * Sets the acknowledge requesting flag
     * 
     * @param _Boolean
     *            The flag value.
     */
    public void setAcknowledgeRequest(boolean _Boolean) {
        if (_Boolean)
            ctrl1 |= 0x0002;
        else
            ctrl1 &= 0xFFFD;
    }

    /**
     * Returns bit 0 of control field 1 inverted. Check this for a L_DATA_CON,
     * to see whether the associated request was processed successfully.
     * 
     * @return the confirmation bit
     */
    public boolean isPositiveConfirmation() {
        if ((ctrl1 & 0x0001) == 1)
            return false;
        return true;
    }

    /**
     * Get the message hop count.
     * 
     * @return hop count
     */
    public byte getHopCount() {
        return (byte) ((ctrl2 & 0x70) >> 4);
    }

    /**
     * Set the message hop count.
     * 
     * @param _hopcount
     *            The routing counter value.
     * @throws IllegalArgumentException
     *             if the value is not within the range 0..7.
     */
    public void setHopCount(byte _hopcount) {
        if ((_hopcount > 7) || (_hopcount < 0))
            throw new IllegalArgumentException("Hop count out of range");

        ctrl2 &= 0x8F;
        ctrl2 |= _hopcount << 4;
    }

    /**
     * Set the message's standard frame bit to 1, and the extended frame format
     * to 0000;
     */
    private void setFrameFormatStandard() {
        ctrl1 |= 0x80;
        ctrl2 &= 0xF0;
    }

    /**
     * Get the length of the message in bytes
     * 
     * @return the message length
     */
    public int getStructLength() {
        return 9 + apci_length;
    }

}
