/*
 * Copyright (C) 2005 Bernhard Erb

 *This program is free software; you can redistribute it and/or
 *modify it under the terms of the GNU General Public License
 *as published by the Free Software Foundation; either version 2
 *of the License, or (at your option) any later version.

 *This program is distributed in the hope that it will be useful,
 *but WITHOUT ANY WARRANTY; without even the implied warranty of
 *MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *GNU General Public License for more details.

 *You should have received a copy of the GNU General Public License
 *along with this program; if not, write to the Free Software
 *Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

/*

 $Log: EIB_Address.java,v $
 Revision 1.12  2005/12/14 22:55:19  gneug
 Eclipse code formatter applied to all source files

 Revision 1.11  2005/12/02 17:35:50  gneug
 String constructor fixed.
 Public constants for address type added.
 byte[] constructor repaired (missing unsigned -> signed conversion).
 Return types for element getters corrected.

 Revision 1.10  2005/12/01 19:49:18  gneug
 Comment around log

 Revision 1.9  2005/12/01 19:45:35  gneug
 Return value of individual address field getters corrected.
 Documentation for new field getters and string constructor added.

 */

package tuwien.auto.eicl.struct.cemi;

import java.util.*;
import org.w3c.dom.*;
import tuwien.auto.eicl.util.*;

/**
 * <p>
 * This class represents a KNX TP1 individual or group address. Such an address
 * consists of a 16-bit value plus one type bit (individual/group). Instances of
 * this class are immutable. They can be created from and exported as strings,
 * byte arrays and XML DOM nodes. Access to the value sub-fields (e.g. zone,
 * line, device) is available as well. Both two- and three-level group addresses
 * are supported.
 * 
 * @author Bernhard Erb, Georg Neugschwandtner
 */

public class EIB_Address {

    private boolean isGroup;

    private short addressHigh;

    private short addressLow;

    /**
     * The XML document node value.
     */
    public static final String XML_EIB_ADDRESS = "eib_address";

    private static final String XML_RAWADDRESS = "address";

    private static final String XML_ISGROUPADDRESS = "is_group";

    public static final boolean GROUP = true;

    public static final boolean INDIVIDUAL = false;

    /**
     * Default (dummy) constructor. Initializes an EIB address instance of 0.0.0
     * (reserved individual address).
     * 
     */
    public EIB_Address() {
        addressHigh = 0;
        addressLow = 0;
        isGroup = false;
    }

    /**
     * 
     * This constructor initializes a new EIB_Address object from a raw 16-bit
     * value. Although it would be possible to pass this value as a parameter of
     * type short, this would be cumbersome since all Java types are signed.
     * Therefore, this parameter is passed as an int.
     * 
     * @param address
     *            the raw address value
     * @param _IsGroupAddress
     *            indicates whether the address object is to be an individual
     *            (false, EIB_Address.INDIVIDUAL) or group address (true,
     *            EIB_Address.GROUP).
     */
    public EIB_Address(int address, boolean _IsGroupAddress) {
        addressHigh = (short) ((address & 0x0000FF00) >> 8);
        addressLow = (short) ((address) & 0x000000FF);
        isGroup = _IsGroupAddress;
    }

    /**
     * This constructor initializes a new EIB_Address object from the first two
     * bytes of a byte array, plus a Boolean value for its kind.
     * 
     * @param _Body
     *            the byte array
     * @param _IsGroupAddress
     *            indicates whether the address object is to be an individual
     *            (false, EIB_Address.INDIVIDUAL) or group address (true,
     *            EIB_Address.GROUP).
     */
    public EIB_Address(byte[] _Body, boolean _IsGroupAddress) {
        /*
         * unsigned to signed conversion: & automatically promotes _Body[n] and
         * 0xFF to ints (with a simple cast, 0xFFs in _Body would end up as
         * addressHigh or addressLow == -1) cf. e.g.,
         * http://darksleep.com/player/JavaAndUnsignedTypes.html
         */
        addressHigh = (short) (_Body[0] & 0xFF);
        addressLow = (short) (_Body[1] & 0xFF);

        isGroup = _IsGroupAddress;
    }

    /**
     * 
     * This constructor is able to initialise the address object from its string
     * representation. The supported formats are:
     * <ul>
     * <li>zone.line.device (individual address, 4-4-8 bits)</li>
     * <li>main/middle/sub (3-level group address, 5-3-8 bits)</li>
     * <li>main/sub (2-level group address, 5-11 bits)</li>
     * </ul>
     * No range checking is performed.
     * 
     * @param _AddressString
     *            the address in string representation
     * @throws EICLException
     *             if another separator than "/" or "." is used or the number of
     *             elements (levels) is wrong
     * 
     */
    public EIB_Address(String _AddressString) throws EICLException {

        StringTokenizer token = null;
        if (_AddressString.indexOf('/') > -1) {
            token = new StringTokenizer(_AddressString, "/");
            isGroup = true;
            if (token.countTokens() == 2) {
                short high = new Short(token.nextToken()).shortValue();
                short low = new Short(token.nextToken()).shortValue();
                addressHigh = (short) (high << 3 | low >>> 8);
                addressLow = (short) (low & 0x00FF);
            } else if (token.countTokens() == 3) {
                addressHigh = (short) (new Short(token.nextToken()).shortValue() << 3 | (new Short(
                        token.nextToken()).shortValue() & 0x0007));
                addressLow = new Short(token.nextToken()).shortValue();
            } else
                throw new EICLException("Wrong group address syntax");
        } else if (_AddressString.indexOf('.') > -1) {
            token = new StringTokenizer(_AddressString, ".");
            isGroup = false;
            if (token.countTokens() == 3) {
                addressHigh = (short) (new Short(token.nextToken()).shortValue() << 4 | (new Short(
                        token.nextToken()).shortValue() & 0x000F));
                addressLow = new Short(token.nextToken()).shortValue();
            } else {
                throw new EICLException("Wrong individual address syntax");
            }
        } else {
            throw new EICLException("Need . or / separator");
        }

    }

    /**
     * Use this constructor to initialise an EIB_Address object from an XML
     * node. The node must have exactly 2 children and must be named
     * "eib_address".
     * 
     * @param _XMLNODE
     *            the XML node
     * @throws EICLException
     *             If the XML node has the wrong format then this exception is
     *             thrown.
     */
    public EIB_Address(Node _XMLNODE) throws EICLException {
        if (!_XMLNODE.getNodeName().equals(XML_EIB_ADDRESS))
            throw new EICLException("Node has to be a " + XML_EIB_ADDRESS);

        NodeList nl = _XMLNODE.getChildNodes();
        if (nl.getLength() != 2)
            throw new EICLException(XML_EIB_ADDRESS + " node must have 2 child nodes");

        for (int i = 0; i < nl.getLength(); i++) {
            Element element = (Element) nl.item(i);
            if (element.getNodeName().equals(XML_RAWADDRESS)) {
                int val = Integer.valueOf(element.getFirstChild().getNodeValue()).intValue();
                addressHigh = (short) ((val & 0x0000FF00) >> 8);
                addressLow = (short) ((val) & 0x000000FF);
            } else if (element.getNodeName().equals(XML_ISGROUPADDRESS))
                isGroup = Boolean.valueOf(element.getFirstChild().getNodeValue()).booleanValue();
        }

    }

    /**
     * Returns the raw address value as a (new) byte array.
     * 
     * @return The address value. The high byte of the address is placed at
     *         index 0. The address type is not included.
     */
    public byte[] toByteArray() {
        byte[] ret = new byte[2];
        ret[0] = (byte) addressHigh;
        ret[1] = (byte) addressLow;
        return ret;
    }

    /**
     * Returns the raw address value as an int (16 bit representation).
     * 
     * @return The address value. The address type is not included.
     */
    public int getRawAddress() {
        return (addressHigh & 0x00FF) << 8 | (addressLow & 0x00FF);
    }

    /**
     * The returned boolean indicates whether this object is a group or
     * individual address.
     * 
     * @return false if this instance is an individual address, true otherwise.
     */
    public boolean isGroupAddress() {
        return isGroup;
    }

    /**
     * Returns the zone number.
     * 
     * @return the zone number (high nibble of the address high byte)
     * @throws IllegalStateException
     *             if this method is invoked on an instance representing a group
     *             address
     */
    public byte getZone() {
        if (isGroup)
            throw new IllegalStateException(
                    "EIBAddress: Operation not appropriate for a group address");
        return (byte) ((addressHigh & 0x00F0) >> 4);
    }

    /**
     * Returns the line number.
     * 
     * @return the line number (low nibble of the address high byte)
     * @throws IllegalStateException
     *             if this method is invoked on an instance representing a group
     *             address
     */
    public byte getLine() {
        if (isGroup)
            throw new IllegalStateException(
                    "EIBAddress: Operation not appropriate for a group address");
        return (byte) ((addressHigh & 0x000F));
    }

    /**
     * Returns the device number.
     * 
     * @return the line number (address low byte)
     * @throws IllegalStateException
     *             if this method is invoked on an instance representing a group
     *             address
     */
    public short getDevice() {
        if (isGroup)
            throw new IllegalStateException(
                    "EIBAddress: Operation not appropriate for a group address");
        return (short) (addressLow & 0x00FF);
    }

    /**
     * Returns the main group value. This is the same for both the two- and
     * three-level group address notations. The most significant bit is actually
     * reserved, but is not specially handled in this class.
     * 
     * @return the main group (5 most significant address bits)
     * @throws IllegalStateException
     *             if this method is invoked on an instance representing an
     *             individual address
     */
    public byte getMainGroup() {
        if (!isGroup)
            throw new IllegalStateException(
                    "EIBAddress: Operation not appropriate for an individual address");
        return (byte) (addressHigh >>> 3);
    }

    /**
     * Returns the middle group value (for the three-level group address
     * notation).
     * 
     * @return the middle group (3 least significant bits of the address high
     *         byte)
     * @throws IllegalStateException
     *             if this method is invoked on an instance representing an
     *             individual address
     */
    public byte getMiddleGroup() {
        if (!isGroup)
            throw new IllegalStateException(
                    "EIBAddress: Operation not appropriate for an individual address");
        return (byte) (addressHigh & 0x0007);
    }

    /**
     * Returns the sub group value (for the two-level group address notation).
     * 
     * @return the sub group (11 least significant bits of the 16-bit address)
     * @throws IllegalStateException
     *             if this method is invoked on an instance representing an
     *             individual address
     */
    public short getSubGroup11() {
        if (!isGroup)
            throw new IllegalStateException(
                    "EIBAddress: Operation not appropriate for an individual address");
        return (short) ((addressHigh & 0x0007) << 8 | addressLow);
    }

    /**
     * Returns the sub group value (for the three-level group address notation).
     * 
     * @return the sub group (address low byte)
     * @throws IllegalStateException
     *             if this method is invoked on an instance representing an
     *             individual address
     */
    public short getSubGroup8() {
        if (!isGroup)
            throw new IllegalStateException(
                    "EIBAddress: Operation not appropriate for an individual address");
        return (short) (addressLow & 0x00FF);
    }

    /**
     * Returns the address as a string. Depending on the address type the "." or
     * "/" (3-level) notation is used.
     * 
     * @return the address string
     */
    public String toString() {
        if (isGroupAddress())
            return getMainGroup() + "/" + getMiddleGroup() + "/" + getSubGroup8();
        return getZone() + "." + getLine() + "." + getDevice();
    }

    /**
     * This method converts the EIB address object into XML representation.
     * 
     * @param doc
     *            a document object
     * @return the eib_address node object
     */
    public Node toXML(Document doc) {
        Element root = doc.createElement(XML_EIB_ADDRESS);

        Element element = doc.createElement(XML_RAWADDRESS);
        element.appendChild(doc.createTextNode(new Integer(addressHigh << 8 | addressLow).toString()));
        root.appendChild(element);

        element = doc.createElement(XML_ISGROUPADDRESS);
        element.appendChild(doc.createTextNode(new Boolean(isGroup).toString()));
        root.appendChild(element);

        return root;
    }

}
