/*
 * Copyright (C) 2005 Bernhard Erb

 *This program is free software; you can redistribute it and/or
 *modify it under the terms of the GNU General Public License
 *as published by the Free Software Foundation; either version 2
 *of the License, or (at your option) any later version.

 *This program is distributed in the hope that it will be useful,
 *but WITHOUT ANY WARRANTY; without even the implied warranty of
 *MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *GNU General Public License for more details.

 *You should have received a copy of the GNU General Public License
 *along with this program; if not, write to the Free Software
 *Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
package tuwien.auto.eicl.struct.eibnetip;

import java.io.*;

import tuwien.auto.eicl.struct.eibnetip.util.EIBNETIP_Constants;
import tuwien.auto.eicl.util.*;

/**
 * <p>
 * This class is the implementation of the standard EIBnet/IP packet structure
 * of version 1.0. This object is useful for creating and parsing EIBnet/IP
 * messages. The message body is not processed, but can be accessed through the
 * getData() method. This can be achieved through other classes in this package.
 * 
 * @see tuwien.auto.eicl.struct.eibnetip.Connect_Request
 * @see tuwien.auto.eicl.struct.eibnetip.Connect_Response
 * @see tuwien.auto.eicl.struct.eibnetip.Connectionstate_Request
 * @see tuwien.auto.eicl.struct.eibnetip.Connectionstate_Response
 * @see tuwien.auto.eicl.struct.eibnetip.Description_Request
 * @see tuwien.auto.eicl.struct.eibnetip.Description_Response
 * @see tuwien.auto.eicl.struct.eibnetip.Disconnect_Request
 * @see tuwien.auto.eicl.struct.eibnetip.Disconnect_Response
 * @see tuwien.auto.eicl.struct.eibnetip.Search_Request
 * @see tuwien.auto.eicl.struct.eibnetip.Search_Response
 * @see tuwien.auto.eicl.struct.eibnetip.CEMI_Connection_Ack
 * @see tuwien.auto.eicl.struct.eibnetip.CEMI_Connection_Request
 * @author Bernhard Erb
 */

public class EIBnetIPPacket {
    /* header */
    short headersize;

    short version;

    int servicetype;

    int totalsize;

    /* body */
    byte[] data;

    /**
     * This constructor initializes a new EIBNET/IP packet object with the
     * relevant header information, service type and total packet size.
     * 
     * @param _ServiceType
     *            the service type.
     * @param _Totalsize
     *            the total packet size.
     */
    public EIBnetIPPacket(int _ServiceType, int _Totalsize) {
        headersize = EIBNETIP_Constants.HEADER_SIZE_10;
        version = EIBNETIP_Constants.EIBNETIP_VERSION_10;
        servicetype = _ServiceType;
        totalsize = _Totalsize;
    }

    /**
     * The constructor parses a new packet object from a byte array. The package
     * information can be accessed through get/set methods.
     * 
     * @param _Packet
     *            the byte array
     * @param _PacketSize
     *            the size
     * @throws EICLException
     *             If something doesn't correspond with the frame format this
     *             exception is thrown.
     */
    public EIBnetIPPacket(byte[] _Packet, int _PacketSize) throws EICLException {
        headersize = _Packet[0];
        version = _Packet[1];

        short high = _Packet[2];
        short low = _Packet[3];
        servicetype = ((high & 0x00FF) << 8 | (low & 0x00FF));

        high = _Packet[4];
        low = _Packet[5];
        totalsize = ((high & 0x00FF) << 8 | (low & 0x00FF));

        // if headersize version or totalsize don't macht throw exception
        if (headersize != EIBNETIP_Constants.HEADER_SIZE_10)
            throw new EICLException("Packet has wrong headersize");

        if (version != EIBNETIP_Constants.EIBNETIP_VERSION_10)
            throw new EICLException("Packet with wrong version");

        if (totalsize != _PacketSize)
            throw new EICLException("Packet with wrong totalsize");

        data = new byte[_PacketSize - EIBNETIP_Constants.HEADER_SIZE_10];
        for (int i = EIBNETIP_Constants.HEADER_SIZE_10; i < _PacketSize; i++) {
            data[i - EIBNETIP_Constants.HEADER_SIZE_10] = _Packet[i];
        }
    }

    /**
     * Returns the service type code as 2 byte value.
     * 
     * @return The service type.
     */
    public int getServiceType() {
        return servicetype;
    }

    /**
     * With this method you can get the byte array representation of this object
     * 
     * @return the byte array
     */
    public byte[] toByteArray() {
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        baos.write((byte) (headersize & 0x00FF));
        baos.write((byte) (version & 0x00FF));
        baos.write((byte) ((servicetype & 0x0000FF00) >> 8));
        baos.write((byte) (servicetype & 0x000000FF));
        baos.write((byte) (totalsize >> 8 & 0x000000FF));
        baos.write((byte) (totalsize & 0x000000FF));
        return baos.toByteArray();
    }

    /**
     * This message returns the message body as byte array which can be parsed
     * using other classes in this package
     * 
     * @return the body
     */
    public byte[] getBody() {
        return data;
    }
}
