/*
 * Copyright (C) 2005 Bernhard Erb

 *This program is free software; you can redistribute it and/or
 *modify it under the terms of the GNU General Public License
 *as published by the Free Software Foundation; either version 2
 *of the License, or (at your option) any later version.

 *This program is distributed in the hope that it will be useful,
 *but WITHOUT ANY WARRANTY; without even the implied warranty of
 *MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *GNU General Public License for more details.

 *You should have received a copy of the GNU General Public License
 *along with this program; if not, write to the Free Software
 *Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

/* $Log: Device_DIB.java,v $
/* Revision 1.8  2005/12/14 22:55:19  gneug
/* Eclipse code formatter applied to all source files
/*
 /* Revision 1.7  2005/12/02 17:25:47  gneug
 /* Removed unused helper byte array in constructor
 /*
 */

package tuwien.auto.eicl.struct.eibnetip.util;

import java.io.*;

import tuwien.auto.eicl.struct.cemi.EIB_Address;

/**
 * 
 * This class implements a representation of the EIBnet/IP device description
 * information block. This structure is used to describe an EIBnet/IP server.
 * All encapsulated information can be accessed through get and set methods.
 * 
 * @author bernhard erb
 * @see tuwien.auto.eicl.struct.eibnetip.util.EIBNETIP_Constants
 * @see tuwien.auto.eicl.struct.eibnetip.Description_Response
 */
public class Device_DIB {
    private short structlength;

    private short descriptiontypecode;

    private short knxmedium;

    private short devicestatus;

    private EIB_Address eibaddress;

    private int projectinstallationindentifier;

    private short[] serialnumber = new short[6];

    private short[] multicastaddress = new short[4];

    private short[] macaddress = new short[6];

    private String name;

    /**
     * This constructor parses a byte array representation of an EIBNET/IP
     * device description information block.
     * 
     * @param _Body
     *            The byte array representing the Device_DIB.
     */
    public Device_DIB(byte[] _Body) {
        ByteArrayInputStream bais = new ByteArrayInputStream(_Body);
        structlength = (short) bais.read();
        descriptiontypecode = (short) bais.read();
        knxmedium = (short) bais.read();
        devicestatus = (short) bais.read();

        byte[] help = new byte[2];
        help[0] = (byte) bais.read();
        help[1] = (byte) bais.read();

        eibaddress = new EIB_Address(help, false);

        short high = (short) bais.read();
        short low = (short) bais.read();
        projectinstallationindentifier = ((high & 0x00FF) << 8 | (low & 0x00FF));

        for (int i = 0; i < serialnumber.length; i++) {
            serialnumber[i] = (short) bais.read();
        }

        for (int i = 0; i < multicastaddress.length; i++) {
            multicastaddress[i] = (short) bais.read();
        }

        for (int i = 0; i < macaddress.length; i++) {
            macaddress[i] = (short) bais.read();
        }

        StringBuffer sbuf = new StringBuffer(30);
        do {
            char c = (char) bais.read();
            if (c == 0)
                break;
            sbuf.append(c);
        } while (sbuf.length() < 30);
        name = sbuf.toString();
    }

    /**
     * Returns the structure length of this message.
     * 
     * @return The structure length
     */
    public short getStructlength() {
        return structlength;
    }

    /**
     * Get the description type code of this message
     * 
     * @return The description type code
     */
    public short getDescriptionTypeCode() {
        return descriptiontypecode;
    }

    /**
     * Retrieve the KNX- medium byte
     * 
     * @return the KNX medium byte
     */
    public short getKnxMedium() {
        return knxmedium;
    }

    /**
     * Get a human readable representation of the KNX Medium type
     * 
     * @return the knx medium string
     */
    public String getKnxMediumString() {

        switch (knxmedium) {
        case EIBNETIP_Constants.TP0:
            return "TP0";
        case EIBNETIP_Constants.TP1:
            return "TP1";
        case EIBNETIP_Constants.PL110:
            return "PL110";
        case EIBNETIP_Constants.PL132:
            return "PL132";
        case EIBNETIP_Constants.RF:
            return "RF";
        default:
            return "No such medium known";
        }
    }

    /**
     * Get the sending device status byte.
     * 
     * @return The status byte
     */
    public short getDeviceStatus() {
        return devicestatus;
    }

    /**
     * Return the sending device individual address.
     * 
     * @return The individual address
     */
    public EIB_Address getEibaddress() {
        return eibaddress;
    }

    /**
     * Get the project and installation identifier as int.
     * 
     * @return The project installation identifier
     */
    public int getProjectInstallationsIdentifier() {
        return projectinstallationindentifier;
    }

    /**
     * Get the device serial number
     * 
     * @return the serial number
     */
    public short[] getSerialNumber() {
        return serialnumber;
    }

    /**
     * Get a device serial number string representation.
     * 
     * @return The device serial number as string.
     */
    public String getSerialNumberString() {
        byte[] help = new byte[serialnumber.length];
        for (int i = 0; i < serialnumber.length; i++) {
            help[i] = (byte) serialnumber[i];
        }
        return new String(help);
    }

    /**
     * Get the device multicast address.
     * 
     * @return Return the multicast address.
     */
    public short[] getMulticastAddress() {
        return multicastaddress;
    }

    /**
     * Get a multicast address string representation.
     * 
     * @return The multicast address string
     */
    public String getMulticastAddressString() {
        return (multicastaddress[0] & 0x00FF) + "." + (multicastaddress[1] & 0x00FF) + "."
                + (multicastaddress[2] & 0x00FF) + "." + (multicastaddress[3] & 0x00FF);
    }

    /**
     * Get the sending point MAC address.
     * 
     * @return The servers interface MACAddress
     */
    public short[] getMACAddress() {
        return macaddress;
    }

    /**
     * Get a human readable MAC Address representation.
     * 
     * @return The MAC address as string.
     */
    public String getMACAddressString() {
        return (macaddress[0] & 0x00FF) + "." + (macaddress[1] & 0x00FF) + "."
                + (macaddress[2] & 0x00FF) + "." + (macaddress[3] & 0x00FF) + "."
                + (macaddress[4] & 0x00FF) + "." + (macaddress[5] & 0x00FF);
    }

    /**
     * Get the device name.
     * 
     * @return The device name.
     */
    public String getName() {
        return name;
    }

}