/*
 * Copyright (C) 2005 Bernhard Erb

 *This program is free software; you can redistribute it and/or
 *modify it under the terms of the GNU General Public License
 *as published by the Free Software Foundation; either version 2
 *of the License, or (at your option) any later version.

 *This program is distributed in the hope that it will be useful,
 *but WITHOUT ANY WARRANTY; without even the implied warranty of
 *MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *GNU General Public License for more details.

 *You should have received a copy of the GNU General Public License
 *along with this program; if not, write to the Free Software
 *Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

/*

 $Log: HPAI.java,v $
 Revision 1.11  2005/12/15 01:10:11  gneug
 Minor correction in Javadoc comment for setLocalIP()

 Revision 1.10  2005/12/14 22:55:19  gneug
 Eclipse code formatter applied to all source files

 Revision 1.9  2005/12/01 20:16:52  gneug
 Documentation for new methods to override automatic IP detection added.


 */

package tuwien.auto.eicl.struct.eibnetip.util;

import java.util.*;
import java.net.*;
import java.io.*;

import tuwien.auto.eicl.util.*;

/**
 * <p>
 * This class is the implementation of the EIBnet/IP Host Protocol Address
 * Information (HPAI) structure. HPAI is used to describe a server/client end
 * point. Its structure varies according to the used underlying protocol. This
 * class is implemented for IPv4.
 * 
 * @author Bernhard Erb
 */

public class HPAI {
    private short structlength;

    private short hostprotocol;

    private byte[] address;

    private int port;

    private static InetAddress local = null;

    /**
     * Creates an HPAI instance from its byte array representation.
     * 
     * @param _Data
     *            The HPAI as byte array.
     * @throws EICLException
     *             If the host protocol in the parsed structure is another than
     *             IPv4.
     */
    public HPAI(byte[] _Data) throws EICLException {
        structlength = _Data[0];
        hostprotocol = _Data[1];

        address = new byte[4];
        address[0] = _Data[2];
        address[1] = _Data[3];
        address[2] = _Data[4];
        address[3] = _Data[5];

        int high = _Data[6] & 0x000000FF;
        int low = _Data[7] & 0x000000FF;
        port = (high << 8 | low);

        if (hostprotocol != EIBNETIP_Constants.IPV4_UDP)
            throw new EICLException("Unsupported protocol found in HPAI");
    }

    /**
     * Allows to retrieve the IP address that is used for the local host.
     * 
     * @return The IP address used by the HPAI(int) constructor.
     */
    public static InetAddress getLocalIP() {
        if (local == null) {
            try {
                local = InetAddress.getLocalHost();
                if (local.isLoopbackAddress()) {
                    Enumeration enumeration = NetworkInterface.getNetworkInterfaces();
                    while (enumeration.hasMoreElements()) {
                        NetworkInterface iface = (NetworkInterface) enumeration.nextElement();
                        Enumeration e = iface.getInetAddresses();
                        while (e.hasMoreElements()) {
                            InetAddress ia = (InetAddress) e.nextElement();
                            if (!ia.isLoopbackAddress() && !ia.isLinkLocalAddress()
                                    && !ia.equals(InetAddress.getByName("127.0.0.1"))) {
                                local = ia;
                            }
                        }
                    }
                }
            } catch (Exception ex) {
                // ignore
            }
        }
        return local;
    }

    /**
     * Allows to set the IP address to be used for the local host. This is only
     * necessary when the address is not correctly detected automatically.
     * 
     * @param ia
     *            The IP address to be used by the HPAI(int) constructor.
     */
    public static void setLocalIP(InetAddress ia) {
        local = ia;
    }

    /**
     * Creates a new HPAI object for the local host. The IP address is
     * determined automatically. Since there does not seem to be a way to access
     * routing information in a portable manner, the automatic detection fails
     * when more than one network interface is installed on the local host. In
     * this case, one of them is chosen at random. The result of the
     * autodetection process can be retrieved at any time via getLocalIP() and
     * can be overridden using the setLocalIP(InetAddress) method.
     * 
     * @param _PortNumber
     *            the local port to be used in the HPAI
     * @throws EICLException
     */
    public HPAI(int _PortNumber) throws EICLException {
        try {
            address = getLocalIP().getAddress();
            structlength = 8;
            hostprotocol = EIBNETIP_Constants.IPV4_UDP;
            port = _PortNumber;
        } catch (Exception ex) {

            throw new EICLException(ex.getMessage());
        }
    }

    /**
     * Converts the HPAI to its byte array representation.
     * 
     * @return the byte array representation of the HPAI
     * @throws EICLException
     *             if the conversion fails
     */
    public byte[] toByteArray() throws EICLException {
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        try {
            baos.write((byte) structlength & 0x00FF);
            baos.write((byte) hostprotocol & 0x00FF);
            baos.write(address);
            baos.write((byte) ((port & 0x0000FF00) >> 8));
            baos.write((byte) (port & 0x000000FF));
        } catch (IOException ex) {
            throw new EICLException(ex.getMessage());
        }
        return baos.toByteArray();
    }

    /**
     * Retrieves the structure size.
     * 
     * @return the structure length
     */
    public short getStructLength() {
        return structlength;
    }

    /**
     * Returns the IP network address in byte format (big endian).
     * 
     * @return the IP address
     */
    public byte[] getAddress() {
        return address;
    }

    /**
     * Offers the IP network address in dot notation as string.
     * 
     * @return The IPv4 addresss as string
     */
    public String getAddressString() {
        String ret = (address[0] & 0x00FF) + "." + (address[1] & 0x00FF) + "."
                + (address[2] & 0x00FF) + "." + (address[3] & 0x00FF);
        return ret;
    }

    /**
     * Returns the IP address and port number as an InetSocketAddress object.
     * 
     * @return The hostdescription as InetSocketaAddress
     */
    public InetSocketAddress getSocketAddress() {
        InetSocketAddress isa = new InetSocketAddress(getAddressString(), getPort());
        return isa;
    }

    /**
     * Offers a human readable host description consisting of IP address and
     * port
     * 
     * @return A human readable host description.
     */
    public String getDescription() {
        String ret = "host " + getAddressString() + " port " + getPort();
        return ret;
    }

    /**
     * Returns the port as integer
     * 
     * @return The port contained in HPAI
     */
    public int getPort() {
        return port;
    }
}
